# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the AOFORCE module from TurboMole"""

import logging
import re
import os
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import AncillaryParser
from turbomoleparser.TurbomoleCommon import RE_FLOAT, RE_FIXED_FLOAT
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")


class AOFORCEparser(object):

    mode_units = {
        # FIXME: add other units for mode energies supported by turbomole
        "cm**(-1)": "inversecm"
    }

    def __init__(self, context, key="aoforce"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.__hessian = None
        self.__modes = None
        self.__energies = None
        self.__energy_unit = None
        self.__activity_infrared = None
        self.__activity_raman = None
        self.__intensities = None
        self.__current_columns = None
        self.__current_row = None
        self.__prefix_width = None
        self.__vibration_files = dict()

    def purge_data(self):
        self.__hessian = None
        self.__modes = None
        self.__energies = None
        self.__activity_infrared = None
        self.__activity_raman = None
        self.__intensities = None
        self.__current_columns = None
        self.__current_row = None
        self.__prefix_width = None
        self.__vibration_files = dict()

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{5,}[^+ ]+",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(r"\s*a o f o r c e - program\s*$",
                    name="header",
                    coverageIgnore=True,
                    subMatchers=[references],
                    endReStr=r"\s*\+-+\+"
                    )

        sub_matchers = [
            header,
            self.__context["geo"].build_qm_geometry_matcher(),
            self.__context["geo"].build_orbital_basis_matcher(),
            self.__context["orbitals"].build_ir_rep_matcher(),
            self.__context["method"].build_dft_functional_matcher(),
            self.__context["method"].build_dftd3_vdw_matcher(),
            self.__build_hessian_matcher(),
            self.__build_normal_mode_file_matcher(),
            self.__build_vibration_spectrum_file_matcher(),
            self.__build_normal_modes_matcher(),
            self.__build_total_energy_matcher(),
        ]

        return self.__context.build_module_matcher("force", sub_matchers, "AOFORCE",
                                                   self.__context["method"].add_default_functional)

    def __build_hessian_file_matcher(self):
        def store_file_name(backend, groups):
            self.__vibration_files["hessian"] = groups[0]
        return SM(r"\s*\*{3}\s*projected\s+hessian\s+written\s+onto\s+\$hessian\s+\(projected\),"
                  r"\s+file=<([^>]+)>\s*\*{3}\s*$",
                  name="hessian file",
                  startReAction=store_file_name
                  )

    def __build_normal_mode_file_matcher(self):
        def store_file_name(backend, groups):
            self.__vibration_files["modes"] = groups[0]
        return SM(r"\s*\*{3}\s*normal\s+modes\s+written\s+onto\s+\$vibrational\s+normal\s+modes,"
                  r"\s+file=<([^>]+)>\s*\*{3}\s*$",
                  name="normal modes file",
                  startReAction=store_file_name
                  )

    def __build_vibration_spectrum_file_matcher(self):
        def store_file_name(backend, groups):
            self.__vibration_files["spectrum"] = groups[0]

        sub = SM("\s*file\s*=\s*<(.+)>\s*\*{3}\s*$",
                 name="spectrum filename",
                 required=True,
                 startReAction=store_file_name)
        return SM(r"\s*\*{3}\s*vibrational\s+spectroscopic\s+data\s+written\s+"
                  r"onto\s*\$vibrational\s+spectrum\s*$",
                  name="spectrum filename",
                  subMatchers=[sub]
                  )

    def __build_hessian_matcher(self):
        re_header = re.compile(r"\s*([0-9]+)\s*[A-z]+")
        re_element = re.compile(r"\s*("+RE_FIXED_FLOAT+")")
        indices = {"x": 0, "y": 1, "z": 2}

        def prepare_data(backend, groups):
            n_atoms = self.__context["geo"].num_atoms()
            self.__hessian = np.ndarray(shape=(n_atoms, n_atoms, 3, 3), dtype=float)
            self.__hessian[:, :, :, :] = 0.0
            self.__current_columns = list()
            self.__current_row = -1

        def process_header(backend, groups):
            match = re_header.match(groups[0])
            del self.__current_columns[:]
            while match:
                self.__current_columns.append(int(match.group(1)) - 1)
                match = re_header.match(groups[0], pos=match.end(0))

        def process_row(backend, groups):
            if groups[0]:
                self.__current_row = int(groups[0]) - 1
            axis1 = indices[groups[1]]
            match = re_element.match(groups[2])
            index = 0
            while match:
                axis2 = index % 3
                atom2 = self.__current_columns[index // 3]
                value = float(match.group(0))
                self.__hessian[self.__current_row, atom2, axis1, axis2] = value
                self.__hessian[atom2, self.__current_row, axis2, axis1] = value
                match = re_element.match(groups[2], pos=match.end(0))
                index += 1

        def write_data(backend, gIndex, section):
            if "hessian" in self.__vibration_files:
                hessian_parser = self.__build_hessian_file_parser()
                dir_name = os.path.dirname(os.path.abspath(self.__context.fName))
                file_name = os.path.normpath(os.path.join(dir_name,
                                                          self.__vibration_files["hessian"]))
                try:
                    with open(file_name) as file_in:
                        hessian_parser.parseFile(file_in)
                except IOError:
                    logger.warning("Could not find auxiliary file '%s' in directory '%s'." % (
                        self.__vibration_files["modes"], dir_name))
            backend.addArrayValues("hessian_matrix", self.__hessian,
                                   self.__context.index_configuration(), unit="hartree * bohr**-2")

        block_row = SM(r"\s*(?:([0-9]+)+\s*[A-z]+\s*)?d([xyz])\s+((?:"+RE_FIXED_FLOAT+"\s*)+)\s*$",
                       name="hessian row",
                       repeats=True,
                       startReAction=process_row
                       )

        block_header = SM(r"\s*ATOM((?:\s+[0-9]+\s*[A-z]+)+)\s*$",
                          name="column header",
                          startReAction=process_header,
                          repeats=True,
                          subMatchers=[
                              SM(r"\s*dx\s+dy\s+dz\s+dx\s+dy\s+dz\s*$", name="column header"),
                              block_row
                          ]
                          )

        return SM(r"\s*CARTESIAN\s+FORCE\s+CONSTANT\s+MATRIX\s+\(hartree/bohr\*\*2\)\s*$",
                  name="Cartesian Hessian",
                  startReAction=prepare_data,
                  subMatchers=[
                      block_header,
                      self.__build_hessian_file_matcher()
                  ],
                  onClose={None: write_data}
                  )

    def __build_total_energy_matcher(self):

        def set_total_energy(backend, groups):
            backend.addRealValue("energy_total", float(groups[0]), unit="hartree")

        scf_etot = SM(r"\s*\*\s*SCF-energy\s+:\s+("+RE_FLOAT+")\s*\*\s*$",
                      name="SCF total energy")
        scf_zp = SM(r"\s*\*\s*SCF\s+\+\s+E\(vib0\)\s+:\s+"
                    r"(?P<energy_current__hartree>"+RE_FLOAT+")\s*\*\s*$",
                    name="SCF total energy",
                    startReAction=set_total_energy)

        return SM(r"\s*\*\s*zero\s+point\s+VIBRATIONAL\s+energy\s+:\s+"
                  r"(?P<energy_zero_point__hartree>"+RE_FLOAT+")\s*Hartree\s*\*\s*$",
                  name="zero point energy",
                  # TODO: define allowed values for ZP-corrections and add output
                  subMatchers=[
                      scf_etot,
                      scf_zp
                  ])

    def __prepare_vibrations__data(self, backend, groups):
        if self.__modes is None:
            n_atoms = self.__context["geo"].num_atoms()
            self.__modes = np.ndarray(shape=(3 * n_atoms, n_atoms, 3), dtype=float)
            self.__modes[:, :, :] = 0.0
            self.__energies = np.ndarray(shape=(3 * n_atoms), dtype=float)
            self.__energies[:] = 0.0
            try:
                self.__energy_unit = self.mode_units[groups[0]]
            except KeyError:
                logger.error("unknown vibrational mode energy unit found: " + groups[0])
                self.__energy_unit = self.mode_units["cm**(-1)"]
            self.__intensities = np.ndarray(shape=(3 * n_atoms), dtype=float)
            self.__intensities[:] = 0.0
            self.__activity_infrared = np.ndarray(shape=(3 * n_atoms), dtype=bool)
            self.__activity_infrared[:] = False
            self.__activity_raman = np.ndarray(shape=(3 * n_atoms), dtype=bool)
            self.__activity_raman[:] = False
            backend.addValue("x_turbomole_vibrations_num_modes", 3 * n_atoms)
            self.__current_columns = list()
            self.__current_row = -1

    def __build_normal_modes_matcher(self):
        re_header = re.compile(r"\s*([0-9]+)")
        re_element = re.compile(r"\s*("+RE_FLOAT+")")
        re_activity = re.compile(r"\s*(-|YES)")
        indices = {"x": 0, "y": 1, "z": 2}

        def process_header(backend, groups):
            match = re_header.match(groups[0])
            del self.__current_columns[:]
            while match:
                self.__current_columns.append(int(match.group(1)) - 1)
                match = re_header.match(groups[0], pos=match.end(0))

        def process_row(backend, groups):
            if groups[0]:
                self.__current_row = int(groups[0]) - 1
            axis1 = indices[groups[1]]
            match = re_element.match(groups[2])
            index = 0
            while match:
                value = float(match.group(0))
                self.__modes[self.__current_columns[index], self.__current_row, axis1] = value
                match = re_element.match(groups[2], pos=match.end(0))
                index += 1

        def extract_frequencies(backend, groups):
            match = re_element.match(groups[0])
            index = 0
            while match:
                value = float(match.group(0))
                self.__energies[self.__current_columns[index]] = value
                match = re_element.match(groups[0], pos=match.end(0))
                index += 1

        def extract_infrared_activity(backend, groups):
            match = re_activity.match(groups[0])
            index = 0
            while match:
                value = match.group(0) == "YES"
                self.__activity_infrared[self.__current_columns[index]] = value
                match = re_activity.match(groups[0], pos=match.end(0))
                index += 1

        def extract_raman_activity(backend, groups):
            match = re_activity.match(groups[0])
            index = 0
            while match:
                value = match.group(0) == "YES"
                self.__activity_raman[self.__current_columns[index]] = value
                match = re_activity.match(groups[0], pos=match.end(0))
                index += 1

        def extract_intensities(backend, groups):
            match = re_element.match(groups[0])
            index = 0
            while match:
                value = float(match.group(0))
                self.__intensities[self.__current_columns[index]] = value
                match = re_element.match(groups[0], pos=match.end(0))
                index += 1

        def write_data(backend, gIndex, section):
            # check if the external files are present and read them if possible (higher precision)
            if "modes" in self.__vibration_files:
                modes_parser = self.__build_normal_modes_file_parser()
                dir_name = os.path.dirname(os.path.abspath(self.__context.fName))
                file_name = os.path.normpath(os.path.join(dir_name,
                                                          self.__vibration_files["modes"]))
                try:
                    with open(file_name) as file_in:
                        modes_parser.parseFile(file_in)
                except IOError:
                    logger.warning("Could not find auxiliary file '%s' in directory '%s'." % (
                        self.__vibration_files["modes"], dir_name))
            if "spectrum" in self.__vibration_files:
                spectrum_parser = self.__build_vibrational_spectrum_file_parser()
                dir_name = os.path.dirname(os.path.abspath(self.__context.fName))
                file_name = os.path.normpath(os.path.join(dir_name,
                                                          self.__vibration_files["spectrum"]))
                try:
                    with open(file_name) as file_in:
                        spectrum_parser.parseFile(file_in)
                except IOError:
                    logger.warning("Could not find auxiliary file '%s' in directory '%s'." % (
                        self.__vibration_files["spectrum"], dir_name))

            backend.addArrayValues("x_turbomole_vibrations_normal_modes", self.__modes,
                                   self.__context.index_configuration(), unit="bohr")
            backend.addArrayValues("x_turbomole_vibrations_mode_energies", self.__energies,
                                   self.__context.index_configuration(), unit=self.__energy_unit)
            backend.addArrayValues("x_turbomole_vibrations_intensities", self.__intensities,
                                   self.__context.index_configuration(), unit="kcal * mole ** -1")
            backend.addArrayValues("x_turbomole_vibrations_infrared_activity",
                                   self.__activity_infrared, self.__context.index_configuration())
            backend.addArrayValues("x_turbomole_vibrations_raman_activity", self.__activity_raman,
                                   self.__context.index_configuration())

        frequencies = SM(r"\s*frequency((?:\s+"+RE_FLOAT+")+)\s*$",
                         name="frequencies",
                         required=True,
                         startReAction=extract_frequencies
                         )
        intensities = SM(r"\s*intensity\s*\(km/mol\)((?:\s+"+RE_FLOAT+")+)\s*$",
                         name="intensities",
                         required=True,
                         startReAction=extract_intensities
                         )
        activity_infrared = SM(r"\s*IR((?:\s+(?:-|YES))+)\s*$",
                               name="IR activity",
                               required=True,
                               startReAction=extract_infrared_activity
                               )
        activity_raman = SM(r"\s*RAMAN((?:\s+(?:-|YES))+)\s*$",
                            name="RAMAN activity",
                            required=True,
                            startReAction=extract_raman_activity
                            )

        block_row = SM(r"\s*(?:([0-9]+)\s+[A-z]+\s+)?([xyz])((?:\s+"+RE_FLOAT+")+)\s*$",
                       name="mode displacements",
                       repeats=True,
                       startReAction=process_row
                       )

        block_header = SM(r"\s*mode((?:\s+[0-9]+)+)\s*$",
                          name="column header",
                          startReAction=process_header,
                          repeats=True,
                          subMatchers=[
                              frequencies,
                              activity_infrared,
                              intensities,
                              activity_raman,
                              block_row
                          ]
                          )

        return SM(r"\s*NORMAL\s+MODES\s+and\s+VIBRATIONAL\s+FREQUENCIES\s+\((.+)\)\s*$",
                  name="vibrational spectrum",
                  startReAction=self.__prepare_vibrations__data,
                  subMatchers=[
                      block_header
                  ],
                  onClose={None: write_data}
                  )

    def __build_hessian_file_parser(self):
        re_element = re.compile(r"\s*("+RE_FLOAT+")")

        def prepare_data(backend, gIndex, section):
            self.__prefix_width = len(str(self.__context["geo"].num_atoms()))+1
            self.__current_columns = 0

        def process_line(backend, groups):
            row_index = int(groups[0][0:self.__prefix_width])
            if self.__current_row != row_index:
                self.__current_columns = 0
                self.__current_row = row_index
            atom_row = (row_index - 1) // 3
            axis_row = (row_index - 1) % 3
            match = re_element.match(groups[1])
            while match:
                atom_col = self.__current_columns // 3
                axis_col = self.__current_columns % 3
                value = float(match.group(0))
                self.__hessian[atom_row, atom_col, axis_row, axis_col] = value
                match = re_element.match(groups[1], pos=match.end(0))
                self.__current_columns += 1

        data = SM(r"(\s*[0-9]+\s*[0-9]+)((?:\s+"+RE_FLOAT+")+)\s*$",
                  name="hessian matrix",
                  startReAction=process_line,
                  repeats=True)
        normal_modes = SM(r"",
                          name="hessian matrix file root",
                          onOpen={None: prepare_data},
                          subMatchers=[
                              SM(r"\$hessian \(projected\)\s*$", name="start", required=True),
                              data,
                              SM(r"\$end\s*$", name="end", required=True)
                          ]
                          )

        return AncillaryParser(fileDescription=normal_modes,
                               parser=self.__context.parser,
                               cachingLevelForMetaName=dict(),
                               superContext=self.__context
                               )

    def __build_normal_modes_file_parser(self):
        re_element = re.compile(r"\s*("+RE_FLOAT+")")

        def prepare_data(backend, gIndex, section):
            self.__prefix_width = len(str(self.__context["geo"].num_atoms()))+1
            self.__current_columns = 0

        def process_line(backend, groups):
            row_index = int(groups[0][0:self.__prefix_width])
            if self.__current_row != row_index:
                self.__current_columns = 0
                self.__current_row = row_index
            atom = (row_index - 1) // 3
            axis = (row_index - 1) % 3
            match = re_element.match(groups[1])
            while match:
                value = float(match.group(0))
                self.__modes[self.__current_columns, atom, axis] = value
                match = re_element.match(groups[1], pos=match.end(0))
                self.__current_columns += 1

        data = SM(r"(\s*[0-9]+\s*[0-9]+)((?:\s+"+RE_FLOAT+")+)\s*$",
                  name="displacements",
                  startReAction=process_line,
                  repeats=True)
        normal_modes = SM(r"",
                          name="normal modes file root",
                          onOpen={None: prepare_data},
                          subMatchers=[
                              SM(r"\$vibrational normal modes\s*$", name="start", required=True),
                              data,
                              SM(r"\$end\s*$", name="end", required=True)
                          ]
                          )

        return AncillaryParser(fileDescription=normal_modes,
                               parser=self.__context.parser,
                               cachingLevelForMetaName=dict(),
                               superContext=self.__context
                               )

    def __build_vibrational_spectrum_file_parser(self):

        def process_header(backend, groups):
            try:
                self.__energy_unit = self.mode_units[groups[0]]
            except KeyError:
                logger.error("unknown vibrational mode energy unit found: " + groups[0])
                self.__energy_unit = self.mode_units["cm**(-1)"]

        def process_line(backend, groups):
            mode_index = int(groups[0]) - 1
            self.__energies[mode_index] = float(groups[1])
            self.__intensities[mode_index] = float(groups[2])
            self.__activity_infrared[mode_index] = groups[3] == "YES"
            self.__activity_raman[mode_index] = groups[4] == "YES"

        header = SM(r"#\s+mode\s+symmetry\s+wave\s+number\s+IR\s+intensity\s+selection\s+rules\s*$",
                    name="header",
                    required=True
                    )
        header_units = SM(r"#\s+(\S+)\s+km/mol\s+IR\s+RAMAN",
                          name="units",
                          required=True,
                          startReAction=process_header
                          )
        data = SM(r"\s*([0-9]+)\s+[a-z'\"]*\s+(" + RE_FLOAT + ")\s+("
                  + RE_FLOAT + ")\s+(-|YES)\s+(-|YES)\s*$",
                  name="vibrational mode",
                  startReAction=process_line,
                  repeats=True)
        normal_modes = SM(r"",
                          name="vibrational spectrum file root",
                          subMatchers=[
                              SM(r"\$vibrational spectrum\s*$", name="start", required=True),
                              header,
                              header_units,
                              data,
                              SM(r"\$end\s*$", name="end", required=True)
                          ]
                          )

        return AncillaryParser(fileDescription=normal_modes,
                               parser=self.__context.parser,
                               cachingLevelForMetaName=dict(),
                               superContext=self.__context
                               )



from builtins import object
# import setup_paths
import numpy as np
import ase.io
import os
import sys

from nomadcore.simple_parser import mainFunction, AncillaryParser, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import wien2kparser.wien2k_parser_struct as wien2k_parser_struct
import wien2kparser.wien2k_parser_in0 as wien2k_parser_in0
import wien2kparser.wien2k_parser_in1c as wien2k_parser_in1c
import wien2kparser.wien2k_parser_in2c as wien2k_parser_in2c
import wien2kparser.wien2k_parser_in1 as wien2k_parser_in1
import wien2kparser.wien2k_parser_in2 as wien2k_parser_in2
import logging

from nomad.parsing.legacy import CoESimpleMatcherParser

################################################################
# This is the parser for the main output file (.scf) of WIEN2k.
################################################################

# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"


class Wien2kContext(object):
    """context for wien2k parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv
        self.rootSecMethodIndex = None
        self.secMethodIndex = None
        self.secSystemIndex = None
        self.scfIterNr = 0

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_x_wien2k_header(self, backend, gIndex, section):
        version_check = section["x_wien2k_version"]
#        riprova = section["x_wien2k_release_date"][0]
#        print("prova=",prova," riprova=",riprova)
        if version_check:
            backend.addValue("program_version",
                             section["x_wien2k_version"][0] + " " +
                             section["x_wien2k_release_date"][0])
        else:
            backend.addValue("program_version", "Before_wien2k11")

    def onOpen_section_system(self, backend, gIndex, section):
        self.secSystemIndex = gIndex

    def onOpen_section_method(self, backend, gIndex, section):

        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".in0"
        if os.path.exists(fName):
            subSuperContext = wien2k_parser_in0.Wien2kIn0Context()
            subParser = AncillaryParser(
                fileDescription = wien2k_parser_in0.buildIn0Matchers(),
                parser = self.parser,
                cachingLevelForMetaName = wien2k_parser_in0.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = subSuperContext)
            with open(fName) as fIn:
                subParser.parseFile(fIn)


        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".in1c"
        if os.path.exists(fName):
            subSuperContext = wien2k_parser_in1c.Wien2kIn1cContext()
            subParser = AncillaryParser(
                fileDescription = wien2k_parser_in1c.buildIn1cMatchers(),
                parser = self.parser,
                cachingLevelForMetaName = wien2k_parser_in1c.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = subSuperContext)
            with open(fName) as fIn:
                subParser.parseFile(fIn)

        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".in2c"
        if os.path.exists(fName):
            subSuperContext = wien2k_parser_in2c.Wien2kIn2cContext()
            subParser = AncillaryParser(
                fileDescription = wien2k_parser_in2c.buildIn2cMatchers(),
                parser = self.parser,
                cachingLevelForMetaName = wien2k_parser_in2c.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = subSuperContext)
            with open(fName) as fIn:
                subParser.parseFile(fIn)

        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".in1"
        if os.path.exists(fName):
            subSuperContext = wien2k_parser_in1.Wien2kIn1Context()
            subParser = AncillaryParser(
                fileDescription = wien2k_parser_in1.buildIn1Matchers(),
                parser = self.parser,
                cachingLevelForMetaName = wien2k_parser_in1.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = subSuperContext)
            with open(fName) as fIn:
                subParser.parseFile(fIn)


        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".in2"
        if os.path.exists(fName):
            subSuperContext = wien2k_parser_in2.Wien2kIn2Context()
            subParser = AncillaryParser(
                fileDescription = wien2k_parser_in2.buildIn2Matchers(),
                parser = self.parser,
                cachingLevelForMetaName = wien2k_parser_in2.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                superContext = subSuperContext)
            with open(fName) as fIn:
                subParser.parseFile(fIn)

        #if self.secMethodIndex is None:
        if self.rootSecMethodIndex is None:
            self.rootSecMethodIndex = gIndex
        self.secMethodIndex = gIndex
#        self.secMethodIndex["single_configuration_to_calculation_method_ref"] = gIndex


    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
       # write number of SCF iterations
        backend.addValue('number_of_scf_iterations', self.scfIterNr)
        # write the references to section_method and section_system
        backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodIndex)
        backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemIndex)


    def onClose_section_system(self, backend, gIndex, section):

        #backend.addValue("smearing_kind", x_fleur_smearing_kind)
        smearing_kind = section['x_wien2k_smearing_kind']
        if smearing_kind is not None:
        #    value = ''
            backend.addValue('x_wien2k_smearing_kind', value)


        smearing_width = section['x_wien2k_smearing_width']
        if smearing_width is not None:
        #    value = ''
            backend.addValue('x_wien2k_smearing_width', value)

        #   atom labels
        atom_labels = section['x_wien2k_atom_name']
        if atom_labels is not None:
           backend.addArrayValues('atom_labels', np.asarray(atom_labels))

        # atom force
        atom_force = []
        for i in ['x', 'y', 'z']:
            api = section['x_wien2k_for_' + i]
            if api is not None:
               atom_force.append(api)
        if atom_force:
            # need to transpose array since its shape is [number_of_atoms,3] in\the metadata
           backend.addArrayValues('atom_forces', np.transpose(np.asarray(atom_force)))

        # Parse the structure file
        mainFile = self.parser.fIn.fIn.name
        fName = mainFile[:-4] + ".struct"
        if os.path.exists(fName):

            # ASE does not support reading file object for WIEN2k structure files.
            try:
                atoms = ase.io.read(fName, format="struct")
            except Exception:
                logging.error("Could not read/parse the WIEN2k structure file.")
            else:
                pos = atoms.get_positions() * 1E-10
                symbols = atoms.get_chemical_symbols()
                cell = atoms.get_cell() * 1E-10
                pbc = atoms.get_pbc()
                backend.addArrayValues('lattice_vectors', cell)
                backend.addArrayValues("configuration_periodic_dimensions", pbc)
                backend.addValue("atom_labels", symbols)
                backend.addArrayValues('atom_positions', pos)

            with open(fName, "r") as fin:
                structSuperContext = wien2k_parser_struct.Wien2kStructContext()
                structParser = AncillaryParser(
                    fileDescription = wien2k_parser_struct.buildStructureMatchers(),
                    parser = self.parser,
                    cachingLevelForMetaName = wien2k_parser_struct.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
                    superContext = structSuperContext)
                structParser.parseFile(fin)

    def onClose_section_scf_iteration(self, backend, gIndex, section):
        #Trigger called when section_scf_iteration is closed.

        # count number of SCF iterations
        self.scfIterNr += 1

# description of the input
mainFileDescription = SM(
    name = 'root',
    weak = True,
    startReStr = "",
    sections   = ['section_run','x_wien2k_header'],
    subMatchers = [
        SM(r"\s*:LABEL[0-9]+: using WIEN2k_(?P<x_wien2k_version>[0-9.]+) \(Release (?P<x_wien2k_release_date>[0-9/.]+)\) in "),
        SM(name = 'newRun',
#           subMatchers=[
#           SM(r"\s*:LABEL[0-9]+: using WIEN2k_(?P<x_wien2k_version>[0-9.]+) \(Release (?P<x_wien2k_release_date>[0-9/.]+)\) in ")
#           ],
           startReStr = r"\s*:ITE[0-9]+:\s*[0-9]+.\s*ITERATION",
           repeats = True,
           required = True,
           forwardMatch = True,
           sections   = ['section_method', 'section_system', 'section_single_configuration_calculation'],
           fixedStartValues={'program_name': 'WIEN2k', 'program_basis_set_type': '(L)APW+lo' }, #, 'program_version': 'Before WIEN2k_11'},
           subMatchers = [
               SM(
                  name = "scf iteration",
                  startReStr = r"\s*:ITE(?P<x_wien2k_iteration_number>[0-9]+):\s*[0-9]*. ITERATION",
                  sections=["section_scf_iteration"],
                  repeats = True,
                  subMatchers=[
                      SM(r":NATO\s*:\s*(?P<x_wien2k_nr_of_independent_atoms>[0-9]+)\s*INDEPENDENT AND\s*(?P<x_wien2k_total_atoms>[0-9]+)\s*TOTAL ATOMS IN UNITCELL"),
                      SM(r"\s*SUBSTANCE: (?P<x_wien2k_system_name>.*)"),
                      SM(r":POT\s*:\s*POTENTIAL OPTION\s*(?P<x_wien2k_potential_option>[0-9]+)"),
                      SM(r":LAT\s*:\s*LATTICE CONSTANTS=\s*(?P<x_wien2k_lattice_const_a>[0-9.]+)\s*(?P<x_wien2k_lattice_const_b>[0-9.]+)\s*(?P<x_wien2k_lattice_const_c>[0-9.]+)"),
                      SM(r":VOL\s*:\s*UNIT CELL VOLUME\s*=\s*(?P<x_wien2k_unit_cell_volume_bohr3>[0-9.]+)"),
                      SM(r":RKM  : MATRIX SIZE (?P<x_wien2k_matrix_size>[0-9]+)\s*LOs:\s*(?P<x_wien2k_LOs>[0-9.]+)\s*RKM=\s*(?P<x_wien2k_rkm>[0-9.]+)\s*WEIGHT=\s*[0-9.]*\s*\w*:"),
                      SM(r":KPT\s*:\s*NUMBER\s*OF\s*K-POINTS:\s*(?P<x_wien2k_nr_kpts>[-+0-9.]+)"),
                      #SM(r":GMA\s*:\s*POTENTIAL\sAND\sCHARGE\sCUT-OFF\s*(?P<x_wien2k_cutoff>[0-9.]+)\s*Ry\*\*[0-9.]+"),
                      SM(r":GMA\s*:\s*POTENTIAL\sAND\sCHARGE\sCUT-OFF\s*(?P<x_wien2k_cutoff>[0-9.]+)\s*Ry\W\W[0-9.]+"),
                      SM(r":GAP\s*:\s*(?P<x_wien2k_ene_gap__rydberg>[-+0-9.]+)\s*Ry\s*=\s*(?P<x_wien2k_ene_gap_eV>[-+0-9.]+)\s*eV\s*.*"),
                      SM(r":NOE\s*:\s*NUMBER\sOF\sELECTRONS\s*=\s*(?P<x_wien2k_noe>[0-9.]+)"),
                      SM(r":FER\s*:\s(\w*\s*)*-\s\w*\W\w*\WM\W*=\s*(?P<x_wien2k_fermi_ene__rydberg>[-+0-9.]+)"),
                      SM(r":GMA\s*:\s*POTENTIAL\sAND\sCHARGE\sCUT-OFF\s*[0-9.]+\s*Ry\W\W[0-9.]+"),
                      SM(r":CHA(?P<x_wien2k_atom_nr>[-+0-9]+):\s*TOTAL\s*\w*\s*CHARGE INSIDE SPHERE\s*(?P<x_wien2k_sphere_nr>[-+0-9]+)\s*=\s*(?P<x_wien2k_tot_val_charge_sphere>[0-9.]+)",repeats = True),
                      SM(r":CHA\s*:\s*TOTAL\s*\w*\s*CHARGE INSIDE\s*\w*\s*CELL\s=\s*(?P<x_wien2k_tot_val_charge_cell>[-+0-9.]+)"),
                      SM(r":MMTOT: TOTAL MAGNETIC MOMENT IN CELL =\s*(?P<x_wien2k_mmtot>[-+0-9.]+)"),
                      SM(r":MMINT: MAGNETIC MOMENT IN INTERSTITIAL =\s*(?P<x_wien2k_mmint>[-+0-9.]+)"),
                      SM(r":MMI001: MAGNETIC MOMENT IN SPHERE 1 =\s*(?P<x_wien2k_mmi001>[-+0-9.]+)"),
                      SM(r":RTO(?P<x_wien2k_atom_nr>[-+0-9]+)\s*:\s*[0-9]+\s*(?P<x_wien2k_density_at_nucleus_valence>[-+0-9.]+)\s*(?P<x_wien2k_density_at_nucleus_semicore>[-+0-9.]+)\s*(?P<x_wien2k_density_at_nucleus_core>[-+0-9.]+)\s*(?P<x_wien2k_density_at_nucleus_tot>[0-9.]+)",repeats = True),
                      SM(r":NTO\s*:\s*\sTOTAL\s*INTERSTITIAL\s*CHARGE=\s*(?P<x_wien2k_tot_int_charge_nm>[-+0-9.]+)"),
                      SM(r":NTO(?P<x_wien2k_atom_nr>[-+0-9]+)[0-9]*:\s*\sTOTAL\s*CHARGE\s*IN\s*SPHERE\s*(?P<x_wien2k_sphere_nr>[-+0-9]+)\s*=\s*(?P<x_wien2k_tot_charge_in_sphere_nm>[-+0-9.]+)",repeats = True),
                      SM(r":DTO(?P<x_wien2k_atom_nr>[-+0-9]+)[0-9]*:\sTOTAL\s*DIFFERENCE\s*CHARGE\W*\w*\s*IN\s*SPHERE\s*(?P<x_wien2k_sphere_nr>[-+0-9]+)\s*=\s*(?P<x_wien2k_tot_diff_charge>[-+0-9.]+)", repeats = True),
                      SM(r":DIS\s*:\s*CHARGE\sDISTANCE\s*\W*[0-9.]+\sfor\satom\s*[0-9]*\sspin\s[0-9]*\W\s*(?P<x_wien2k_charge_distance>[0-9.]+)"),
                      SM(r":CTO\s*:\s*\sTOTAL\s*INTERSTITIAL\s*CHARGE=\s*(?P<x_wien2k_tot_int_charge>[-+0-9.]+)"),
                      SM(r":CTO(?P<x_wien2k_atom_nr>[-+0-9]+)[0-9]*:\s*\sTOTAL\s*CHARGE\s*IN\s*SPHERE\s*(?P<x_wien2k_sphere_nr>[-+0-9]+)\s*=\s*(?P<x_wien2k_tot_charge_in_sphere>[-+0-9.]+)",repeats = True),
#                      SM(r":NEC(?P<x_wien2k_necnr>[-+0-9]+)\s*:\s*NUCLEAR AND ELECTRONIC CHARGE\s*(?P<x_wien2k_nuclear_charge>[-+0-9.]+)\s*(?P<x_wien2k_electronic_charge>[0-9.]+)",repeats = True),
                      SM(r":ENE\s*:\s*\W*\w*\W*\s*TOTAL\s*ENERGY\s*IN\s*Ry\s*=\s*(?P<x_wien2k_energy_total__rydberg>[-+0-9.]+)"),
                      SM(r":FOR[0-9]*:\s*(?P<x_wien2k_atom_nr>[0-9]+).ATOM\s*(?P<x_wien2k_for_abs>[0-9.]+)\s*(?P<x_wien2k_for_x>[-++0-9.]+)\s*(?P<x_wien2k_for_y>[-+0-9.]+)\s*(?P<x_wien2k_for_z>[-+0-9.]+)\s*partial\sforces", repeats = True),
                      SM(r":FGL[0-9]*:\s*(?P<x_wien2k_atom_nr>[0-9]+).ATOM\s*(?P<x_wien2k_for_x_gl>[-+0-9.]+)\s*(?P<x_wien2k_for_y_gl>[-+0-9.]+)\s*(?P<x_wien2k_for_z_gl>[-+0-9.]+)\s*partial\sforces", repeats = True)
                  ]
              )
           ]
       )
    ])


class Wien2kParser(CoESimpleMatcherParser):

    def metainfo_env(self):
        from .metainfo import m_env
        return m_env

    def create_super_context(self):
        return Wien2kContext()

    def create_simple_matcher(self):
        return mainFileDescription

    def create_parser_description(self):
        return {
            "name": "Wien2k",
            "version": "1.0"
        }

    def create_caching_levels(self):
        return {
            "XC_functional_name": CachingLevel.ForwardAndCache,
            "energy_total": CachingLevel.ForwardAndCache
        }



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

from nomad.metainfo import (
    MSection, Package, Quantity, SubSection, SectionProxy, Section,
    Datetime, JSON)


m_package = Package(name='common_experimental')


class Experiment(MSection):
    '''
    The root section for all (meta)data that belongs to a single experiment.
    '''
    m_def = Section(validate=False)

    experiment_summary = Quantity(
        type=str, description='A descriptive summary of the content of the experiment.')

    experiment_location = SubSection(sub_section=SectionProxy('Location'))

    experiment_publish_time = Quantity(
        type=Datetime, description='The datetime when this experiment was published.')

    experiment_time = Quantity(
        type=Datetime, description='The datetime of the beginning of the experiment.')

    experiment_end_time = Quantity(
        type=Datetime, description='The datetime of the experiment end.')

    raw_metadata = Quantity(
        type=JSON, description='The whole or partial metadata in its original source JSON format.')

    section_data = SubSection(sub_section=SectionProxy('Data'))

    section_method = SubSection(sub_section=SectionProxy('Method'))

    section_sample = SubSection(sub_section=SectionProxy('Sample'))


class Location(MSection):
    m_def = Section(validate=False)

    address = Quantity(
        type=str, description='''
        The address where the experiment took place, format 'Country, City, Street'
        ''')

    institution = Quantity(
        type=str, description='''
        Name of the institution hosting the experimental facility (e.g. in full or an
        acronym).
        ''')

    facility = Quantity(
        type=str, description='''
        Name of the experimental facility (e.g. in full or an acronym).
        ''')


class Data(MSection):
    '''
    This section contains information about the stored data.
    '''
    m_def = Section(validate=False)

    repository_name = Quantity(
        type=str, description='The name of the repository, where the data is stored.')

    repository_url = Quantity(
        type=str, description='An URL to the repository, where the data is stored.')

    preview_url = Quantity(
        type=str, description='An URL to an image file that contains a preview.')

    entry_repository_url = Quantity(
        type=str, description='An URL to the entry on the repository, where the data is stored.')


class Method(MSection):
    '''
    This section contains information about the applied experimental method.
    '''
    m_def = Section(validate=False)

    data_type = Quantity(
        type=str, description='Name of the type of data that the experiment produced.')

    method_name = Quantity(
        type=str, description='Full name of the experimental method in use')

    method_abbreviation = Quantity(
        type=str,
        description='Abbreviated name (i.e. acronym) of the experimental method')

    probing_method = Quantity(
        type=str, description='The probing method used')

    instrument_description = Quantity(
        type=str, description='A description of the instrumentation used for the experiment.')


class Sample(MSection):
    '''
    The section for all sample related (meta)data that was used in the experiment.
    '''
    m_def = Section(validate=False)

    sample_description = Quantity(
        type=str, description='Description of the sample used in the experiment.')

    sample_id = Quantity(
        type=str, description='Identification number or signatures of the sample used.')

    sample_state = Quantity(
        type=str, description='The physical state of the sample.')

    sample_temperature = Quantity(
        type=np.dtype(np.float64), unit='kelvin',
        description='The temperature of the sample during the experiment.')

    sample_microstructure = Quantity(
        type=str, description='The sample microstructure.')

    sample_constituents = Quantity(
        type=str, description='The constituents.')

    section_material = SubSection(sub_section=SectionProxy('Material'))


class Material(MSection):
    ''' This section describes a sample's material. '''
    m_def = Section(validate=False)

    chemical_formula = Quantity(
        type=str, description='The chemical formula that describes the sample.')

    chemical_name = Quantity(
        type=str, description='The chemical name that describes the sample.')

    atom_labels = Quantity(
        type=str, shape=['number_of_elements'],
        description='Atom labels for distinct elements in the sample.')

    number_of_elements = Quantity(
        type=int, derived=lambda m: len(m.atom_labels) if m.atom_labels else 0,
        description='Number of distinct chemical elements in the sample.')

    space_group = Quantity(
        type=int, description='Space group of the sample compound (if crystalline).')


m_package.__init_metainfo__()



import unittest

from lark.exceptions import VisitError

from optimade.filterparser import LarkParser, ParserError
from optimade.filtertransformers.mongo import MongoTransformer
from optimade.server.mappers import BaseResourceMapper


class TestMongoTransformer(unittest.TestCase):
    version = (0, 10, 1)
    variant = "default"

    def setUp(self):
        p = LarkParser(version=self.version, variant=self.variant)
        t = MongoTransformer()
        self.transform = lambda inp: t.transform(p.parse(inp))

    def test_empty(self):
        self.assertIsNone(self.transform(" "))

    def test_property_names(self):
        self.assertEqual(self.transform("band_gap = 1"), {"band_gap": {"$eq": 1}})
        self.assertEqual(
            self.transform("cell_length_a = 1"), {"cell_length_a": {"$eq": 1}}
        )
        self.assertEqual(self.transform("cell_volume = 1"), {"cell_volume": {"$eq": 1}})

        with self.assertRaises(ParserError):
            self.transform("0_kvak IS KNOWN")  # starts with a number

        with self.assertRaises(ParserError):
            self.transform('"foo bar" IS KNOWN')  # contains space; contains quotes

        with self.assertRaises(ParserError):
            self.transform("BadLuck IS KNOWN")  # contains upper-case letters

        # database-provider-specific prefixes
        self.assertEqual(
            self.transform("_exmpl_formula_sum = 1"), {"_exmpl_formula_sum": {"$eq": 1}}
        )
        self.assertEqual(
            self.transform("_exmpl_band_gap = 1"), {"_exmpl_band_gap": {"$eq": 1}}
        )

        # Nested property names
        self.assertEqual(
            self.transform("identifier1.identifierd2 = 42"),
            {"identifier1.identifierd2": {"$eq": 42}},
        )

    def test_string_values(self):
        self.assertEqual(
            self.transform('author="Sąžininga Žąsis"'),
            {"author": {"$eq": "Sąžininga Žąsis"}},
        )
        self.assertEqual(
            self.transform('field = "!#$%&\'() * +, -./:; <= > ? @[] ^ `{|}~ % "'),
            {"field": {"$eq": "!#$%&'() * +, -./:; <= > ? @[] ^ `{|}~ % "}},
        )

    def test_number_values(self):
        self.assertEqual(self.transform("a = 12345"), {"a": {"$eq": 12345}})
        self.assertEqual(self.transform("b = +12"), {"b": {"$eq": 12}})
        self.assertEqual(self.transform("c = -34"), {"c": {"$eq": -34}})
        self.assertEqual(self.transform("d = 1.2"), {"d": {"$eq": 1.2}})
        self.assertEqual(self.transform("e = .2E7"), {"e": {"$eq": 2000000.0}})
        self.assertEqual(self.transform("f = -.2E+7"), {"f": {"$eq": -2000000.0}})
        self.assertEqual(self.transform("g = +10.01E-10"), {"g": {"$eq": 1.001e-09}})
        self.assertEqual(self.transform("h = 6.03e23"), {"h": {"$eq": 6.03e23}})
        self.assertEqual(self.transform("i = .1E1"), {"i": {"$eq": 1.0}})
        self.assertEqual(self.transform("j = -.1e1"), {"j": {"$eq": -1.0}})
        self.assertEqual(self.transform("k = 1.e-12"), {"k": {"$eq": 1e-12}})
        self.assertEqual(self.transform("l = -.1e-12"), {"l": {"$eq": -1e-13}})
        self.assertEqual(
            self.transform("m = 1000000000.E1000000000"), {"m": {"$eq": float("inf")}}
        )

        with self.assertRaises(ParserError):
            self.transform("number=1.234D12")
        with self.assertRaises(ParserError):
            self.transform("number=.e1")
        with self.assertRaises(ParserError):
            self.transform("number= -.E1")
        with self.assertRaises(ParserError):
            self.transform("number=+.E2")
        with self.assertRaises(ParserError):
            self.transform("number=1.23E+++")
        with self.assertRaises(ParserError):
            self.transform("number=+-123")
        with self.assertRaises(ParserError):
            self.transform("number=0.0.1")

    def test_simple_comparisons(self):
        self.assertEqual(self.transform("a<3"), {"a": {"$lt": 3}})
        self.assertEqual(self.transform("a<=3"), {"a": {"$lte": 3}})
        self.assertEqual(self.transform("a>3"), {"a": {"$gt": 3}})
        self.assertEqual(self.transform("a>=3"), {"a": {"$gte": 3}})
        self.assertEqual(self.transform("a=3"), {"a": {"$eq": 3}})
        self.assertEqual(self.transform("a!=3"), {"a": {"$ne": 3}})

    def test_id(self):
        self.assertEqual(self.transform('id="example/1"'), {"id": {"$eq": "example/1"}})
        self.assertEqual(
            self.transform('"example/1" = id'), {"id": {"$eq": "example/1"}}
        )
        self.assertEqual(
            self.transform('id="test/2" OR "example/1" = id'),
            {"$or": [{"id": {"$eq": "test/2"}}, {"id": {"$eq": "example/1"}}]},
        )

    def test_operators(self):
        # Basic boolean operations
        # TODO: {"a": {"$not": {"$lt": 3}}} can be simplified to {"a": {"$gte": 3}}
        self.assertEqual(self.transform("NOT a<3"), {"a": {"$not": {"$lt": 3}}})

        # TODO: {'$not': {'$eq': 'Ti'}} can be simplified to {'$ne': 'Ti'}
        self.assertEqual(
            self.transform(
                "NOT ( "
                'chemical_formula_hill = "Al" AND chemical_formula_anonymous = "A" OR '
                'chemical_formula_anonymous = "H2O" AND NOT chemical_formula_hill = "Ti" '
                ")"
            ),
            {
                "$nor": [
                    {
                        "$and": [
                            {"chemical_formula_hill": {"$eq": "Al"}},
                            {"chemical_formula_anonymous": {"$eq": "A"}},
                        ]
                    },
                    {
                        "$and": [
                            {"chemical_formula_anonymous": {"$eq": "H2O"}},
                            {"chemical_formula_hill": {"$not": {"$eq": "Ti"}}},
                        ]
                    },
                ]
            },
        )

        # Numeric and String comparisons
        self.assertEqual(self.transform("nelements > 3"), {"nelements": {"$gt": 3}})
        self.assertEqual(
            self.transform(
                'chemical_formula_hill = "H2O" AND chemical_formula_anonymous != "AB"'
            ),
            {
                "$and": [
                    {"chemical_formula_hill": {"$eq": "H2O"}},
                    {"chemical_formula_anonymous": {"$ne": "AB"}},
                ]
            },
        )
        self.assertEqual(
            self.transform(
                "_exmpl_aax <= +.1e8 OR nelements >= 10 AND "
                'NOT ( _exmpl_x != "Some string" OR NOT _exmpl_a = 7)'
            ),
            {
                "$or": [
                    {"_exmpl_aax": {"$lte": 10000000.0}},
                    {
                        "$and": [
                            {"nelements": {"$gte": 10}},
                            {
                                "$nor": [
                                    {"_exmpl_x": {"$ne": "Some string"}},
                                    {"_exmpl_a": {"$not": {"$eq": 7}}},
                                ]
                            },
                        ]
                    },
                ]
            },
        )
        self.assertEqual(
            self.transform('_exmpl_spacegroup="P2"'),
            {"_exmpl_spacegroup": {"$eq": "P2"}},
        )
        self.assertEqual(
            self.transform("_exmpl_cell_volume<100.0"),
            {"_exmpl_cell_volume": {"$lt": 100.0}},
        )
        self.assertEqual(
            self.transform("_exmpl_bandgap > 5.0 AND _exmpl_molecular_weight < 350"),
            {
                "$and": [
                    {"_exmpl_bandgap": {"$gt": 5.0}},
                    {"_exmpl_molecular_weight": {"$lt": 350}},
                ]
            },
        )
        self.assertEqual(
            self.transform(
                '_exmpl_melting_point<300 AND nelements=4 AND elements="Si,O2"'
            ),
            {
                "$and": [
                    {"_exmpl_melting_point": {"$lt": 300}},
                    {"nelements": {"$eq": 4}},
                    {"elements": {"$eq": "Si,O2"}},
                ]
            },
        )
        self.assertEqual(
            self.transform("_exmpl_some_string_property = 42"),
            {"_exmpl_some_string_property": {"$eq": 42}},
        )
        self.assertEqual(self.transform("5 < _exmpl_a"), {"_exmpl_a": {"$gt": 5}})

        self.assertEqual(
            self.transform("a<5 AND b=0"),
            {"$and": [{"a": {"$lt": 5}}, {"b": {"$eq": 0}}]},
        )
        self.assertEqual(
            self.transform("a >= 8 OR a<5 AND b>=8"),
            {
                "$or": [
                    {"a": {"$gte": 8}},
                    {"$and": [{"a": {"$lt": 5}}, {"b": {"$gte": 8}}]},
                ]
            },
        )

        # OPTIONAL
        # self.assertEqual(self.transform("((NOT (_exmpl_a>_exmpl_b)) AND _exmpl_x>0)"), {})

        self.assertEqual(
            self.transform("NOT (a>1 AND b>1)"),
            {"$and": [{"a": {"$not": {"$gt": 1}}}, {"b": {"$not": {"$gt": 1}}}]},
        )

        self.assertEqual(
            self.transform("NOT (a>1 AND b>1 OR c>1)"),
            {
                "$nor": [
                    {"$and": [{"a": {"$gt": 1}}, {"b": {"$gt": 1}}]},
                    {"c": {"$gt": 1}},
                ]
            },
        )

        self.assertEqual(
            self.transform("NOT (a>1 AND ( b>1 OR c>1 ))"),
            {
                "$and": [
                    {"a": {"$not": {"$gt": 1}}},
                    {"$nor": [{"b": {"$gt": 1}}, {"c": {"$gt": 1}}]},
                ]
            },
        )

        self.assertEqual(
            self.transform("NOT (a>1 AND ( b>1 OR (c>1 AND d>1 ) ))"),
            {
                "$and": [
                    {"a": {"$not": {"$gt": 1}}},
                    {
                        "$nor": [
                            {"b": {"$gt": 1}},
                            {"$and": [{"c": {"$gt": 1}}, {"d": {"$gt": 1}}]},
                        ]
                    },
                ]
            },
        )

        self.assertEqual(
            self.transform(
                'elements HAS "Ag" AND NOT ( elements HAS "Ir" AND elements HAS "Ac" )'
            ),
            {
                "$and": [
                    {"elements": {"$in": ["Ag"]}},
                    {
                        "$and": [
                            {"elements": {"$not": {"$in": ["Ir"]}}},
                            {"elements": {"$not": {"$in": ["Ac"]}}},
                        ]
                    },
                ]
            },
        )

        self.assertEqual(self.transform("5 < 7"), {7: {"$gt": 5}})

        with self.assertRaises(VisitError):
            self.transform('"some string" > "some other string"')

    def test_filtering_on_relationships(self):
        """ Test the nested properties with special names
        like "structures", "references" etc. are applied
        to the relationships field.

        """

        self.assertEqual(
            self.transform('references.id HAS "dummy/2019"'),
            {"relationships.references.data.id": {"$in": ["dummy/2019"]}},
        )

        self.assertEqual(
            self.transform('structures.id HAS ANY "dummy/2019", "dijkstra1968"'),
            {
                "relationships.structures.data.id": {
                    "$in": ["dummy/2019", "dijkstra1968"]
                }
            },
        )

        self.assertEqual(
            self.transform('structures.id HAS ALL "dummy/2019", "dijkstra1968"'),
            {
                "relationships.structures.data.id": {
                    "$all": ["dummy/2019", "dijkstra1968"]
                }
            },
        )

        self.assertEqual(
            self.transform('structures.id HAS ONLY "dummy/2019"'),
            {
                "$and": [
                    {"relationships.structures.data": {"$size": 1}},
                    {"relationships.structures.data.id": {"$all": ["dummy/2019"]}},
                ]
            },
        )

        self.assertEqual(
            self.transform(
                'structures.id HAS ONLY "dummy/2019" AND structures.id HAS "dummy/2019"'
            ),
            {
                "$and": [
                    {
                        "$and": [
                            {"relationships.structures.data": {"$size": 1}},
                            {
                                "relationships.structures.data.id": {
                                    "$all": ["dummy/2019"]
                                }
                            },
                        ]
                    },
                    {"relationships.structures.data.id": {"$in": ["dummy/2019"]}},
                ],
            },
        )

    def test_not_implemented(self):
        """ Test that list properties that are currently not implemented
        give a sensible response.

        """
        # NOTE: Lark catches underlying filtertransformer exceptions and
        # raises VisitErrors, most of these actually correspond to NotImplementedError
        with self.assertRaises(VisitError):
            try:
                self.transform("list HAS < 3")
            except Exception as exc:
                self.assertTrue("not implemented" in str(exc))
                raise exc

        with self.assertRaises(VisitError):
            try:
                self.transform("list HAS ALL < 3, > 3")
            except Exception as exc:
                self.assertTrue("not implemented" in str(exc))
                raise exc

        with self.assertRaises(VisitError):
            try:
                self.transform("list HAS ANY > 3, < 6")
            except Exception as exc:
                self.assertTrue("not implemented" in str(exc))
                raise exc

        self.assertEqual(self.transform("list LENGTH 3"), {"list": {"$size": 3}})

        with self.assertRaises(VisitError):
            self.transform("list:list HAS >=2:<=5")

        with self.assertRaises(VisitError):
            self.transform(
                'elements:_exmpl_element_counts HAS "H":6 AND elements:_exmpl_element_counts '
                'HAS ALL "H":6,"He":7 AND elements:_exmpl_element_counts HAS ONLY "H":6 AND '
                'elements:_exmpl_element_counts HAS ANY "H":6,"He":7 AND '
                'elements:_exmpl_element_counts HAS ONLY "H":6,"He":7'
            )

        with self.assertRaises(VisitError):
            self.transform(
                "_exmpl_element_counts HAS < 3 AND _exmpl_element_counts "
                "HAS ANY > 3, = 6, 4, != 8"
            )

        with self.assertRaises(VisitError):
            self.transform(
                "elements:_exmpl_element_counts:_exmpl_element_weights "
                'HAS ANY > 3:"He":>55.3 , = 6:>"Ti":<37.6 , 8:<"Ga":0'
            )

        self.assertEqual(
            self.transform("list LENGTH > 3"), {"list.4": {"$exists": True}}
        )

    def test_list_length_aliases(self):
        from optimade.server.mappers import StructureMapper

        transformer = MongoTransformer(mapper=StructureMapper())
        parser = LarkParser(version=self.version, variant=self.variant)

        self.assertEqual(
            transformer.transform(parser.parse("elements LENGTH 3")), {"nelements": 3}
        )

        self.assertEqual(
            transformer.transform(
                parser.parse('elements HAS "Li" AND elements LENGTH = 3')
            ),
            {"$and": [{"elements": {"$in": ["Li"]}}, {"nelements": 3}]},
        )

        self.assertEqual(
            transformer.transform(parser.parse("elements LENGTH > 3")),
            {"nelements": {"$gt": 3}},
        )
        self.assertEqual(
            transformer.transform(parser.parse("elements LENGTH < 3")),
            {"nelements": {"$lt": 3}},
        )
        self.assertEqual(
            transformer.transform(parser.parse("elements LENGTH = 3")), {"nelements": 3}
        )
        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH <= 3")),
            {"nsites": {"$lte": 3}},
        )
        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH >= 3")),
            {"nsites": {"$gte": 3}},
        )

    def test_unaliased_length_operator(self):
        self.assertEqual(
            self.transform("cartesian_site_positions LENGTH <= 3"),
            {"cartesian_site_positions.4": {"$exists": False}},
        )
        self.assertEqual(
            self.transform("cartesian_site_positions LENGTH < 3"),
            {"cartesian_site_positions.3": {"$exists": False}},
        )
        self.assertEqual(
            self.transform("cartesian_site_positions LENGTH 3"),
            {"cartesian_site_positions": {"$size": 3}},
        )
        self.assertEqual(
            self.transform("cartesian_site_positions LENGTH >= 10"),
            {"cartesian_site_positions.10": {"$exists": True}},
        )

        self.assertEqual(
            self.transform("cartesian_site_positions LENGTH > 10"),
            {"cartesian_site_positions.11": {"$exists": True}},
        )

    def test_aliased_length_operator(self):
        from optimade.server.mappers import StructureMapper

        class MyMapper(StructureMapper):
            ALIASES = (("elements", "my_elements"), ("nelements", "nelem"))
            LENGTH_ALIASES = (
                ("chemsys", "nelements"),
                ("cartesian_site_positions", "nsites"),
                ("elements", "nelements"),
            )
            PROVIDER_FIELDS = ("chemsys",)

        transformer = MongoTransformer(mapper=MyMapper())
        parser = LarkParser(version=self.version, variant=self.variant)

        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH <= 3")),
            {"nsites": {"$lte": 3}},
        )
        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH < 3")),
            {"nsites": {"$lt": 3}},
        )
        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH 3")),
            {"nsites": 3},
        )
        self.assertEqual(
            transformer.transform(parser.parse("cartesian_site_positions LENGTH 3")),
            {"nsites": 3},
        )
        self.assertEqual(
            transformer.transform(
                parser.parse("cartesian_site_positions LENGTH >= 10")
            ),
            {"nsites": {"$gte": 10}},
        )

        self.assertEqual(
            transformer.transform(parser.parse("structure_features LENGTH > 10")),
            {"structure_features.11": {"$exists": True}},
        )

        self.assertEqual(
            transformer.transform(parser.parse("nsites LENGTH > 10")),
            {"nsites.11": {"$exists": True}},
        )

        self.assertEqual(
            transformer.transform(parser.parse("elements LENGTH 3")), {"nelem": 3},
        )

        self.assertEqual(
            transformer.transform(parser.parse('elements HAS "Ag"')),
            {"my_elements": {"$in": ["Ag"]}},
        )

        self.assertEqual(
            transformer.transform(parser.parse("chemsys LENGTH 3")), {"nelem": 3},
        )

    def test_aliases(self):
        """ Test that valid aliases are allowed, but do not effect
        r-values.

        """

        class MyStructureMapper(BaseResourceMapper):
            ALIASES = (
                ("elements", "my_elements"),
                ("A", "D"),
                ("B", "E"),
                ("C", "F"),
                ("_exmpl_nested_field", "nested_field"),
            )

        mapper = MyStructureMapper()
        t = MongoTransformer(mapper=mapper)

        self.assertEqual(mapper.alias_for("elements"), "my_elements")

        test_filter = {"elements": {"$in": ["A", "B", "C"]}}
        self.assertEqual(
            t.postprocess(test_filter), {"my_elements": {"$in": ["A", "B", "C"]}},
        )
        test_filter = {"$and": [{"elements": {"$in": ["A", "B", "C"]}}]}
        self.assertEqual(
            t.postprocess(test_filter),
            {"$and": [{"my_elements": {"$in": ["A", "B", "C"]}}]},
        )
        test_filter = {"elements": "A"}
        self.assertEqual(t.postprocess(test_filter), {"my_elements": "A"})
        test_filter = ["A", "B", "C"]
        self.assertEqual(t.postprocess(test_filter), ["A", "B", "C"])

        test_filter = ["A", "elements", "C"]
        self.assertEqual(t.postprocess(test_filter), ["A", "elements", "C"])

        test_filter = {"_exmpl_nested_field.sub_property": {"$gt": 1234.5}}
        self.assertEqual(
            t.postprocess(test_filter), {"nested_field.sub_property": {"$gt": 1234.5}}
        )

        test_filter = {"_exmpl_nested_field.sub_property.x": {"$exists": False}}
        self.assertEqual(
            t.postprocess(test_filter),
            {"nested_field.sub_property.x": {"$exists": False}},
        )

    def test_list_properties(self):
        """ Test the HAS ALL, ANY and optional ONLY queries.

        """
        self.assertEqual(
            self.transform('elements HAS ONLY "H","He","Ga","Ta"'),
            {"elements": {"$all": ["H", "He", "Ga", "Ta"], "$size": 4}},
        )

        self.assertEqual(
            self.transform('elements HAS ANY "H","He","Ga","Ta"'),
            {"elements": {"$in": ["H", "He", "Ga", "Ta"]}},
        )

        self.assertEqual(
            self.transform('elements HAS ALL "H","He","Ga","Ta"'),
            {"elements": {"$all": ["H", "He", "Ga", "Ta"]}},
        )

        self.assertEqual(
            self.transform(
                'elements HAS "H" AND elements HAS ALL "H","He","Ga","Ta" AND elements HAS '
                'ONLY "H","He","Ga","Ta" AND elements HAS ANY "H", "He", "Ga", "Ta"'
            ),
            {
                "$and": [
                    {"elements": {"$in": ["H"]}},
                    {"elements": {"$all": ["H", "He", "Ga", "Ta"]}},
                    {"elements": {"$all": ["H", "He", "Ga", "Ta"], "$size": 4}},
                    {"elements": {"$in": ["H", "He", "Ga", "Ta"]}},
                ]
            },
        )

    def test_known_properties(self):
        #  Filtering on Properties with unknown value
        self.assertEqual(
            self.transform("chemical_formula_anonymous IS UNKNOWN"),
            {
                "$or": [
                    {"chemical_formula_anonymous": {"$exists": False}},
                    {"chemical_formula_anonymous": {"$eq": None}},
                ]
            },
        )
        self.assertEqual(
            self.transform("chemical_formula_anonymous IS KNOWN"),
            {
                "$and": [
                    {"chemical_formula_anonymous": {"$exists": True}},
                    {"chemical_formula_anonymous": {"$ne": None}},
                ]
            },
        )
        self.assertEqual(
            self.transform("NOT chemical_formula_anonymous IS UNKNOWN"),
            {
                "$and": [
                    {"chemical_formula_anonymous": {"$exists": True}},
                    {"chemical_formula_anonymous": {"$ne": None}},
                ]
            },
        )
        self.assertEqual(
            self.transform("NOT chemical_formula_anonymous IS KNOWN"),
            {
                "$or": [
                    {"chemical_formula_anonymous": {"$exists": False}},
                    {"chemical_formula_anonymous": {"$eq": None}},
                ]
            },
        )

        self.assertEqual(
            self.transform(
                "chemical_formula_hill IS KNOWN AND NOT chemical_formula_anonymous IS UNKNOWN"
            ),
            {
                "$and": [
                    {
                        "$and": [
                            {"chemical_formula_hill": {"$exists": True}},
                            {"chemical_formula_hill": {"$ne": None}},
                        ]
                    },
                    {
                        "$and": [
                            {"chemical_formula_anonymous": {"$exists": True}},
                            {"chemical_formula_anonymous": {"$ne": None}},
                        ]
                    },
                ]
            },
        )

        self.assertEqual(
            self.transform(
                "chemical_formula_hill IS KNOWN AND chemical_formula_anonymous IS UNKNOWN"
            ),
            {
                "$and": [
                    {
                        "$and": [
                            {"chemical_formula_hill": {"$exists": True}},
                            {"chemical_formula_hill": {"$ne": None}},
                        ]
                    },
                    {
                        "$or": [
                            {"chemical_formula_anonymous": {"$exists": False}},
                            {"chemical_formula_anonymous": {"$eq": None}},
                        ]
                    },
                ]
            },
        )

    def test_precedence(self):
        self.assertEqual(
            self.transform('NOT a > b OR c = 100 AND f = "C2 H6"'),
            {
                "$or": [
                    {"a": {"$not": {"$gt": "b"}}},
                    {"$and": [{"c": {"$eq": 100}}, {"f": {"$eq": "C2 H6"}}]},
                ]
            },
        )
        self.assertEqual(
            self.transform('NOT a > b OR c = 100 AND f = "C2 H6"'),
            self.transform('(NOT (a > b)) OR ( (c = 100) AND (f = "C2 H6") )'),
        )
        self.assertEqual(
            self.transform("a >= 0 AND NOT b < c OR c = 0"),
            self.transform("((a >= 0) AND (NOT (b < c))) OR (c = 0)"),
        )

    def test_special_cases(self):
        self.assertEqual(self.transform("te < st"), {"te": {"$lt": "st"}})
        self.assertEqual(
            self.transform('spacegroup="P2"'), {"spacegroup": {"$eq": "P2"}}
        )
        self.assertEqual(
            self.transform("_cod_cell_volume<100.0"),
            {"_cod_cell_volume": {"$lt": 100.0}},
        )
        self.assertEqual(
            self.transform("_mp_bandgap > 5.0 AND _cod_molecular_weight < 350"),
            {
                "$and": [
                    {"_mp_bandgap": {"$gt": 5.0}},
                    {"_cod_molecular_weight": {"$lt": 350}},
                ]
            },
        )
        self.assertEqual(
            self.transform(
                '_cod_melting_point<300 AND nelements=4 AND elements="Si,O2"'
            ),
            {
                "$and": [
                    {"_cod_melting_point": {"$lt": 300}},
                    {"nelements": {"$eq": 4}},
                    {"elements": {"$eq": "Si,O2"}},
                ]
            },
        )
        self.assertEqual(self.transform("key=value"), {"key": {"$eq": "value"}})
        self.assertEqual(
            self.transform('author=" someone "'), {"author": {"$eq": " someone "}}
        )
        self.assertEqual(self.transform("notice=val"), {"notice": {"$eq": "val"}})
        self.assertEqual(
            self.transform("NOTice=val"), {"ice": {"$not": {"$eq": "val"}}}
        )
        self.assertEqual(
            self.transform(
                "number=0.ANDnumber=.0ANDnumber=0.0ANDnumber=+0AND_n_u_m_b_e_r_=-0AND"
                "number=0e1ANDnumber=0e-1ANDnumber=0e+1"
            ),
            {
                "$and": [
                    {"number": {"$eq": 0.0}},
                    {"number": {"$eq": 0.0}},
                    {"number": {"$eq": 0.0}},
                    {"number": {"$eq": 0}},
                    {"_n_u_m_b_e_r_": {"$eq": 0}},
                    {"number": {"$eq": 0.0}},
                    {"number": {"$eq": 0.0}},
                    {"number": {"$eq": 0.0}},
                ]
            },
        )



import os
from pathlib import Path
import re
import tempfile
import unittest


class TestSetup(unittest.TestCase):
    """Test ./setup.py"""

    def test_distributions_package(self):
        """Make sure a distribution has all the needed package data"""
        package_root = Path(__file__).parent.parent.resolve()

        number_of_grammar_files = len(
            list(package_root.joinpath("optimade/grammar").rglob("*.lark"))
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            current_dir = Path(temp_dir)
            file_output = current_dir.joinpath("sdist.out")

            os.system(
                f"python {package_root.joinpath('setup.py')} sdist --dry-run > {file_output}"
            )

            with open(file_output, "r") as file_:
                # Removing new-line-chars via slice
                lines = [_[:-1] for _ in file_.readlines()]

            present = {
                r"\.lark": False,
                r"index_links\.json": False,
                r"test_structures\.json": False,
                r"test_references\.json": False,
                r"test_links\.json": False,
                r"filters\.txt": False,
                r"optional_filters\.txt": False,
                r"landing_page\.html": False,
            }
            count = 0
            for line in lines:
                for requirement in present:
                    if re.findall(requirement, line):
                        if requirement == r"\.lark":
                            count += 1
                            if count == number_of_grammar_files:
                                present[requirement] = True
                            else:
                                present[requirement] = False
                        else:
                            present[requirement] = True

            for requirement, is_present in present.items():
                if requirement == r"\.lark":
                    msg = f"{requirement} files NOT found. Expected {number_of_grammar_files}, found instead {count}."
                else:
                    msg = f"{requirement} file NOT found."

                self.assertTrue(is_present, msg=msg)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import os
import sys
import operator
import numpy as np


from .LammpsCommon import get_metaInfo
from .LammpsCommon import converter as lconverter
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import mainFunction
import nomadcore.ActivateLogging
from contextlib import contextmanager


from re import escape as esc




class sformat(object):
    float = r"[-+]?\d*\.\d+"
    int = r"[-+]?\d+"

nformat = {
    'float': r"[-+]?\d*\.\d+",
    'int': r"[-+]?\d+"
}


# import numpy as np
# import setup_paths
# import logging, os, re, sys
# import nomadcore.ActivateLogging

############################################################
# This is the parser for the input file of LAMMPS.
############################################################

logger = logging.getLogger(name="nomad.LammpsDataParser")


@contextmanager # SECTIONS ARE CLOSED AUTOMATICALLY
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)



class LammpsTrjParserContext(object):

    """Context for parsing LAMMPS input file.

    Attributes:
        # dos_energies: Stores parsed energies.
        # dos_values: Stores parsed DOS values.

    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """

    def __init__(self, converter, dump_style, writeMetaData=True):
        """Args:
            writeMetaData: Determines if metadata is written or stored in class attributes.
        """
        self.parser = None
        self.converter = converter
        self.dump_style = dump_style
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        # self.dos_energies = None
        # self.dos_values = None

        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()


    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
        """
        # self.masses = {}
        self.pbcBool = []

        self.simulationCell = []
        self.atomPositionScaled = []
        self.atomPositionScaledBool = []
        self.atomPosition = []
        self.atomPositionBool = []
        self.atomPositionWrapped = []
        self.atomPositionWrappedBool = []
        self.imageFlagIndex = []
        self.imageFlagIndexBool = []
        self.atomVelocity = []

        pass





    def onClose_section_run(self, backend, gIndex, section):

        pass

    def onClose_section_system(self, backend, gIndex, section):

        p = backend.superBackend
        o = open_section

        toMass     = self.converter.ratioMass
        toDistance = self.converter.ratioDistance
        toTime     = self.converter.ratioTime
        toEnergy   = self.converter.ratioEnergy
        toVelocity = self.converter.ratioVelocity
        toForce    = self.converter.ratioForce
        toTorque   = self.converter.ratioTorque
        toTemp     = self.converter.ratioTemp
        toPress    = self.converter.ratioPress
        toDVisc    = self.converter.ratioDVisc
        toCharge   = self.converter.ratioCharge
        toDipole   = self.converter.ratioDipole
        toEField   = self.converter.ratioEField
        toDensity  = self.converter.ratioDensity

        with o(p, 'section_system'):


            box_bound = section["x_lammps_trj_box_bound_store"][0]
            boundary = list(map(lambda x: x == "pp", box_bound.split()))

            self.pbcBool.append(boundary)


            # Calculating simulation cell vectors frame by frame
            simulationCelldata = []
            for line in section['x_lammps_trj_box_bounds_store']:
                simulationCelldata.append([float(x) for x in line.split()])


            xx = [ simulationCelldata[0][0] - simulationCelldata[0][1], 0, 0 ]
            yy = [ 0, simulationCelldata[1][0] - simulationCelldata[1][1], 0 ]
            zz = [ 0, 0, simulationCelldata[2][0] - simulationCelldata[2][1] ]
            simulationCellstore = [xx, yy, zz]

            self.simulationCell.append(simulationCellstore)
            temp_simulation_cell = [ [ dim*toDistance for dim in box ] for box in simulationCellstore ]
            p.addArrayValues('simulation_cell', np.array(temp_simulation_cell))

            variables = section['x_lammps_trj_variables_store'][0]
            variables = variables.split()

            frameAtomInfo = []
            for line in section["x_lammps_trj_atoms_store"]:
                frameAtomInfo.append(line.split())

            if 'id' in variables:
                idInd = variables.index('id')
                frameAtomInfo.sort(key=lambda x: int(x[idInd]))


            isAtomPosition = False
            if 'x' in variables and 'y' in variables and 'z' in variables:     # if true, unwrapped coord are dumped
                isAtomPosition = True

            self.atomPositionBool.append(isAtomPosition)



            isAtomPositionScaled = False
            if 'xs' in variables and 'ys' in variables and 'zs' in variables:  # if true, scaled coord are dumped
                isAtomPositionScaled = True

            self.atomPositionScaledBool.append(isAtomPositionScaled)

            isImageFlagIndex = False
            if 'ix' in variables and 'iy' in variables and 'iz' in variables:  # if true, image flag indexes are dumped
                isImageFlagIndex = True

            self.imageFlagIndexBool.append(isImageFlagIndex)
            # self.imageFlagIndex = imageFlagIndex


            # Atom velocities
            isAtomVelocity = False
            if 'vx' in variables and 'vy' in variables and 'vz' in variables:  # if true, scaled coord are dumped
                isAtomVelocity = True

            # Atom forces
            isAtomForces = False
            if 'fx' in variables and 'fy' in variables and 'fz' in variables:  # if true, scaled coord are dumped
                isAtomForces = True

            # Atom position (unwrapped)
            if isAtomPosition == True:

                xInd = variables.index('x')
                yInd = variables.index('y')
                zInd = variables.index('z')

                store_single =[]
                frame = section["x_lammps_trj_atoms_store"]
                for line in frame:
                    line = line.split()
                    if isImageFlagIndex:
                        ixInd = variables.index('ix')
                        iyInd = variables.index('iy')
                        izInd = variables.index('iz')

                        store = [float(line[xInd]) + int(line[ixInd]) * np.linalg.norm(simulationCellstore[0]),
                                 float(line[yInd]) + int(line[iyInd]) * np.linalg.norm(simulationCellstore[1]),
                                 float(line[zInd]) + int(line[izInd]) * np.linalg.norm(simulationCellstore[2])]
                    else:
                        store = [float(line[xInd]), float(line[yInd]), float(line[zInd])]

                    store_single.append(store)

                self.atomPosition.append(store_single)

                temp_atom_positions = [ [ crd*toDistance for crd in atom ] for atom in store_single ]
                p.addArrayValues('atom_positions', np.asarray(temp_atom_positions))
                # p.addArrayValues('atom_labels', np.asarray(atomAtLabel))

            # else:
            #     atomPosition = []
            #     atomPositionBool = False


            # Atom position (scaled) [0, 1]
            if isAtomPositionScaled == True:

                xsInd = variables.index('xs')
                ysInd = variables.index('ys')
                zsInd = variables.index('zs')

                atomPositionScaled = []

                store_single_scaled =[]

                frame = section["x_lammps_trj_atoms_store"]
                for line in frame:
                    line = line.split()
                    store = [float(line[xsInd]), float(line[ysInd]), float(line[zsInd])]
                    store_single_scaled.append(store)

                self.atomPositionScaled.append(store_single_scaled)



            # Atom position (converted from scaled positions)

                store = []
                store_single = []
                for atom in store_single_scaled:

                    x = atom[0] * np.linalg.norm(simulationCellstore[0]) + np.min(simulationCellstore[0])
                    y = atom[1] * np.linalg.norm(simulationCellstore[1]) + np.min(simulationCellstore[1])
                    z = atom[2] * np.linalg.norm(simulationCellstore[2]) + np.min(simulationCellstore[2])
                    store = [x, y, z]
                    store_single.append(store)

                self.atomPosition.append(store_single)

                temp_atom_positions = [ [ crd*toDistance for crd in atom ] for atom in store_single ]
                p.addArrayValues('atom_positions', np.asarray(temp_atom_positions))

            # else:
            #     atomPositionScaled = []
            #     atomPositionScaledBool = False
            #     atomPosition = []
            #     atomPositionBool = False


            # Atom velocities
            if isAtomVelocity == True:

                vxInd = variables.index('vx')
                vyInd = variables.index('vy')
                vzInd = variables.index('vz')

                # atomVelocity = []

                store_single_velocity =[]

                frame = section["x_lammps_trj_atoms_store"]
                for line in frame:
                    line = line.split()
                    store = [float(line[vxInd]), float(line[vyInd]), float(line[vzInd])]
                    store_single_velocity.append(store)

                self.atomVelocity.append(store_single_velocity)
                temp_atom_velocities = [ [ vi*toVelocity for vi in atom ] for atom in store_single_velocity ]
                p.addArrayValues('atom_velocities', np.asarray(temp_atom_velocities))

            # Atom Forces
            if isAtomForces == True:

                fxInd = variables.index('fx')
                fyInd = variables.index('fy')
                fzInd = variables.index('fz')

                # atomVelocity = []

                store_single_forces =[]

                frame = section["x_lammps_trj_atoms_store"]
                for line in frame:
                    line = line.split()
                    store = [float(line[fxInd]), float(line[fyInd]), float(line[fzInd])]
                    store_single_forces.append(store)

                self.atomVelocity.append(store_single_forces)




                # atomForce = []
                # for frame in frameAtomInfo:
                #     store_1 = []
                #     for line in frame:
                #
                #         if fxInd and fyInd and fzInd:
                #             store = [ float(line[fxInd]), float(line[fyInd]), float(line[fzInd]) ]
                #             store_1.append(store)
                #
                #     atomForce.append(store_1)

            # self.atomPositionWrapped = atomPositionWrapped
            # self.atomPositionWrappedBool  = atomPositionWrappedBool



            # return (simulationCell, atomPositionScaled, atomPositionScaledBool, atomPosition, atomPositionBool,
            #             atomPositionWrapped, atomPositionWrappedBool , imageFlagIndex, imageFlagIndexBool )




                #
            # xx = [ header[5][0] - header[5][1], 0, 0 ]
            # yy = [ 0, header[6][0] - header[6][1], 0 ]
            # zz = [ 0, 0, header[7][0] - header[7][1] ]
            # store = [xx, yy, zz]

        # self.simulationCell.append()

        pass




def build_LammpsTrjFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the TRJ file of Lammps.



    Returns:
       SimpleMatcher that parses TRJ file of Lammps.
    """
    # Define the output parsing tree for this version
    r_float = "[-+]?\d*\.\d+"  # Regex for a floating point value
    r_int = "[-+]?\d+"         # Regex for an integer
    r_file = "[0-9a-zA-Z_.]+"
    r_word = "[\w]+"
    r_words = "[\w\s]+"

    timestep = SM(
        name='lammps-trj-timestep',
        startReStr=r"ITEM:\sTIMESTEP",
        # sections=['section_topology'],
        subMatchers=[
            SM(
                r"(?P<x_lammps_trj_timestep_store>{0})".format(r_int)
            )
        ])

    numberofatoms = SM (
        name = "lampps-trj-number-of-atoms",
        startReStr=r"ITEM:\sNUMBER\sOF\sATOMS",
        subMatchers=[
            SM(
                r"(?P<x_lammps_trj_number_of_atoms_store>{0})".format(r_int)
            )
        ],
    )

    boxbound = SM (
        name = "lampps-trj-box-bound",
        startReStr=r"ITEM:\sBOX\sBOUNDS\s(?P<x_lammps_trj_box_bound_store>{0}\s{0}\s{0})".format(r_word),
        subMatchers=[
            SM(r"(?P<x_lammps_trj_box_bounds_store>[\s\d.eE-]+)".format(r_int, r_float), repeats=True)
        ],
    )


    atoms = SM (
        name = "lampps-trj-box-bound",
        startReStr=r"ITEM:\sATOMS\s(?P<x_lammps_trj_variables_store>{0})+".format(r_words),
        subMatchers=[
            SM(r"(?P<x_lammps_trj_atoms_store>[\s\deE.-]+)".format(r_int, r_float), repeats=True)
        ],
    )

    return SM(
        name = 'root',
        startReStr = r"",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
            SM(
                name = 'root-timestep',
                startReStr = r"ITEM:\sTIMESTEP",
                sections = ['section_system'],
                forwardMatch=True,
                repeats=True,
                subMatchers=[
                    timestep,
                    numberofatoms,
                    boxbound,
                    atoms
                ]
            )
        ]
    )




def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
        'section_run': CachingLvl,
        'section_topology': CachingLvl,
    }

    # Set all temporarly metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.endswith('_store'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(converter = lconverter(), dump_style="atom", CachingLvl = CachingLevel.Forward):
    """Main function.

    Set up everything for the parsing of the FHI-aims DOS file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get dos file description
    LammpsTrjSimpleMatcher = build_LammpsTrjFileSimpleMatcher()

    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../../../nomad-meta-info/meta_info/nomad_meta_info/lammps.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)

    # set parser info
    parserInfo = {'name': 'lammps-data-parser', 'version': '1.0'}

    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)

    # start parsing
    mainFunction(
        mainFileDescription=LammpsTrjSimpleMatcher,
        metaInfoEnv=metaInfoEnv,
        parserInfo=parserInfo,
        cachingLevelForMetaName=cachingLevelForMetaName,
        superContext=LammpsTrjParserContext(converter, dump_style)
    )


if __name__ == "__main__":
    main(dump_style="custom", CachingLvl=CachingLevel.Forward)



# Copyright 2016-2018 Fawzi Mohamed, Danio Brambila
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# phonopy parser based on the original work of Joerg Mayer on phonopy-FHI-aims

import numpy as np
import math
import os
import logging
from fnmatch import fnmatch
from phonopy import Phonopy

from phonopyparser.FHIaims import read_aims_output

class Control:
    def __init__(self, file=None):
        self.phonon = {}
        self.phonon["supercell"] = []
        self.phonon["displacement"] = 0.001
        self.phonon["symmetry_thresh"] = 1E-6
        self.phonon["frequency_unit"] = "cm^-1"
        self.phonon["nac"] = {}
        if file is None:
            self.file = "control.in"
        else:
            self.file = file
        self.read_phonon()

    def read_phonon(self):
        f = open(self.file, 'r')
        try:
            for line in f:
                if not line:
                    break
                fields = line.split()
                nfields = len(fields)
                if (nfields > 2) and (fields[0] == "phonon"):
                    if (fields[1] == "supercell"):
                        if (nfields >= 11):
                            s = map(int, fields[2:11])
                            s = list(s)
                            Smat = np.array(s).reshape(3, 3)
                        elif (nfields >= 5):
                            s = map(int, fields[2:5])
                            s = list(s)
                            Smat = np.diag(s)
                        else:
                            raise Exception("invalid supercell")
                        det_Smat = np.linalg.det(Smat)
                        if (det_Smat == 0):
                            raise Exception("supercell matrix is not invertible")
                        # this consistency check is not strictly necessary, since int function above used to transform the input
                        # already throws an exception when decimal numbers are encountered
                        # keep for consistency (and future changes to that spot?) nevertheless...
                        elif (abs(det_Smat - round(det_Smat)) > 1.0e-6):
                            raise Exception("determinant of supercell differs from integer by more than numerical tolerance of 1.0e-6")
                        self.phonon["supercell"] = s
                    if (fields[1] == "displacement"):
                        self.phonon["displacement"] = float(fields[2])
                    if (fields[1] == "symmetry_thresh"):
                        self.phonon["symmetry_thresh"] = float(fields[2])
                    if (fields[1] == "frequency_unit"):
                        self.phonon["frequency_unit"] = fields[2]
                    if (fields[1] == "nac") and (len(fields) >= 4):
                        if (len(fields) >= 7):
                            delta = tuple(list(map(float, fields[4:7])))
                        else:
                            delta = (0.1, 0.1, 0.1)
                        if (delta[0] == 0.0) and (delta[1] == 0.0) and (delta[2] == 0.0):
                            raise Exception("evaluation of frequencies with non-analytic corrections must be shifted by delta away from Gamma")
                        parameters = {
                            "file": fields[2], "method": fields[3].lower(), "delta": delta}
                        self.phonon["nac"].update(parameters)

        except Exception as e:
            raise(e)

        # supercell is mandatory for all what follows
        if not self.phonon["supercell"]:
            raise Exception("no supercell specified in %s" % self.file)
        f.close()


def clean_position(scaled_positions):
    scaled_positions = list(scaled_positions)
    for sp in range(len(scaled_positions)):
        for i in range(len(scaled_positions[sp])):
            if np.float(np.round(scaled_positions[sp][i], 7)) >= 1:
                scaled_positions[sp][i] -= 1.0
            elif scaled_positions[sp][i] <= -1e-5:
                scaled_positions[sp][i] += 1.0
    scaled_positions = np.array(scaled_positions)

    return scaled_positions


def read_forces_aims(cell_obj, supercell_matrix, displacement, sym, tol=1e-6, logger=None):
    if logger is None:
        logger = logging

    phonopy_obj = Phonopy(cell_obj, supercell_matrix, symprec=sym)
    phonopy_obj.generate_displacements(distance=displacement)
    supercells = phonopy_obj.get_supercells_with_displacements()
    directories = []
    digits = int(math.ceil(math.log(len(supercells) + 1, 10))) + 1
    for i in range(len(supercells)):
        directories.append(("phonopy-FHI-aims-displacement-%0" + str(digits) + "d") % (i + 1))
    set_of_forces = []
    Relative_Path = []
    for directory, supercell in zip(directories, supercells):
        aims_out = os.path.join(directory, directory + ".out")
        if not os.path.isfile(aims_out):
            logger.warn("!!! file not found: %s" % aims_out)
            cwd = os.getcwd()
            con_list = os.listdir(cwd)
            check_var = False
            for name in con_list:
                if fnmatch(name, '*.out'):
                    aims_out = '%s/%s' % (directory, name)
                    logger.warn(
                        "Your file seems to have a wrong name proceeding with %s" % aims_out
                    )
                    check_var = True
                    break
            if not check_var:
                raise Exception("No phonon calculations found")
            os.chdir("../")
        Relative_Path.append(aims_out)
        supercell_calculated = read_aims_output(aims_out)
        if (
            (supercell_calculated.get_number_of_atoms() == supercell.get_number_of_atoms()) and
            (supercell_calculated.get_atomic_numbers() == supercell.get_atomic_numbers()).all() and
            (abs(supercell_calculated.get_positions() - supercell.get_positions()) < tol).all() and
            (abs(supercell_calculated.get_cell() - supercell.get_cell()) < tol).all()):
            # read_aims_output reads in forces from FHI-aims output as list structure,
            # but further processing below requires numpy array
            forces = np.array(supercell_calculated.get_forces())
            drift_force = forces.sum(axis=0)
            for force in forces:
                force -= drift_force / forces.shape[0]
            set_of_forces.append(forces)
        elif (
            (supercell_calculated.get_number_of_atoms() == supercell.get_number_of_atoms()) and
            (supercell_calculated.get_atomic_numbers() == supercell.get_atomic_numbers()).all() and
            (abs(clean_position(supercell_calculated.get_scaled_positions()) - clean_position(supercell.get_scaled_positions())) < tol).all() and
            (abs(supercell_calculated.get_cell() - supercell.get_cell()) < tol).all()):
            logger.warn("!!! there seems to be a rounding error")
            forces = np.array(supercell_calculated.get_forces())
            drift_force = forces.sum(axis=0)
            for force in forces:
                force -= drift_force / forces.shape[0]
            set_of_forces.append(forces)
        else:
            raise Exception("calculated varies from expected supercell in FHI-aims output %s" % aims_out)

    return set_of_forces, phonopy_obj, Relative_Path



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
The optimade implementation of NOMAD.
'''

from flask import Blueprint
from flask_restplus import Api

from .api import blueprint, url, api
from .structures import Structure, StructuresInfo, StructureList
from .calculations import CalculationList, Calculation, CalculationInfo
from .infolinks import Info, Links
from .index import Info
from .filterparser import parse_filter



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys, os, os.path
baseDir = os.path.dirname(os.path.abspath(__file__))
commonDir = os.path.normpath(os.path.join(baseDir,"../../../../python-common/common/python"))

if not commonDir in sys.path:
    sys.path.insert(0, commonDir)



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ase import data
import re
from atkparser.physical_quantities import eV, Angstrom, Bohr, Kelvin, Hour, Hartree

All = 'All'
Automatic = 'Automatic'
HamiltonianVariable = 'HamiltonianVariable'
SphericalSymmetric = 'SphericalSymmetric'


LDA = type('LDA', (object,), {'PZ': 'LDA.PZ',
                              'PW': 'LDA.PW',
                              'RPA': 'LDA.RPA'})()

GGA = type('GGA', (object,), {'PBE': 'GGA.PBE',
                              'RPBE': 'GGA.RPBE',
                              'PW91': 'GGA.PW91',
                              'PBES': 'GGA.PBES'})()

ptable = {name: symbol for symbol, name in zip(data.chemical_symbols,
                                               data.atomic_names)}
PeriodicTable = type('PeriodicTable', (object,), ptable)()

Preconditioner = type('Preconditioner', (object,), {'Off': 'Off',

                                                    'On': 'On'})


# Populate with dummy classes to hold hold variables (assumes kwargs!)
# Otherwise we have to build all of them with something like
# class LCAOCalculator(object):
#     def __init__(self, basis_set=None, ...)
#
# is easily done, but a bit more work at the moment...
#
def init(self, *args, **kwargs):
    #if len(args)>0:
    #    print(*args)
    #assert len(args) == 0
    self.args = args
    for key, value in kwargs.items():
        setattr(self, key, value)


clss = ['LCAOCalculator', 'BasisSet', 'ConfinedOrbital', 'CheckpointHandler',
        'ParallelParameters', 'AlgorithmParameters', 'DiagonalizationSolver',
        'FastFourierSolver', 'IterationControlParameters',
        'NumericalAccuracyParameters', 'MonkhorstPackGrid',
        'NormConservingPseudoPotential', 'AnalyticalSplit', 'ConfinedOrbital',
        'PolarizationOrbital', 'PulayMixer', 'HydrogenOrbital']

for cls in clss:
    code = cls + ' = type("' + cls + '", (object,)' + ', {"__init__": init})'
    exec(code)


def parse_calculator(fd, calcname):
    """calc: the configuratio the calcualtor refers to
       The name of the calculator in the nc-file is
       conf_calculator, fx BulkConfiguration_gID000_calculator
    """
    code = fd.variables[calcname].data[:].copy()
    code = code.tostring().decode("utf-8")
    if 0:
        print(code)
    #s = re.search('\s*(?P<name>[0-9a-zA-Z_]+)\s*=\s*LCAOCalculator\(', code)
    #name = s.group('name')
    exec(code)
    for obj in locals().values():
        if isinstance(obj, LCAOCalculator):
            return obj
    assert 0, 'No calculator found'

if __name__ == '__main__':
    from scipy.io.netcdf import netcdf_file
    fd = netcdf_file('Water.nc', 'r')
    calc = parse_calculator(fd, 'BulkConfiguration_gID000_calculator')
    print(dir(calc))



from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from .QboxCommon import get_metaInfo
import logging, os, re, sys



#################################################################################
# This is the parser for the *.xml file of qbox to read the geometry information.
# Only needed when 'load *.xml' is used in the *.r file.
###################################################################################

logger = logging.getLogger("nomad.QboxXMLParser")

class QboxXMLParserContext(object):

    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        self.fName = fInName


    def onClose_section_system(self, backend, gIndex, section):
        """Trigger called when section_system is closed.
        Writes atomic positions, atom labels and lattice vectors.
        """
        # keep track of the latest system description section
        self.secSystemDescriptionIndex = gIndex

       #------1.atom_positions
        atom_pos = []
        for i in ['x', 'y', 'z']:
            api = section['x_qbox_geometry_atom_positions_' + i]
            if api is not None:
               atom_pos.append(api)
        if atom_pos:
            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
           backend.addArrayValues('atom_positions', np.transpose(np.asarray(atom_pos)))

        #------2.atom labels
        atom_labels = section['x_qbox_geometry_atom_labels']
        if atom_labels is not None:
           backend.addArrayValues('atom_labels', np.asarray(atom_labels))





################################################################################
######################  MAIN PARSER STARTS HERE  ###############################
################################################################################

def build_QboxXMLFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.xml file of qbox.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.xml file of qbox.
    """

    geometrySubMatcher = SM(name = 'Geometry',
        startReStr = r"\s*<atomset>",
        sections = ['section_system'],
        subMatchers = [
        SM (startReStr = r"\s*<unit_cell\s*",
            subMatchers = [
            SM (r"\s*[a-z]=\"\s*(?P<x_qbox_geometry_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<x_qbox_geometry_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<x_qbox_geometry_lattice_vector_z__bohr>[-+0-9.]+)\s*\"", repeats = True)
            ]),
        SM (startReStr = r"\s*<atom\s+name=\"(?P<x_qbox_geometry_atom_labels>[a-zA-Z0-9]+)\"",
            subMatchers = [
            SM (r"\s*<position>\s+(?P<x_qbox_geometry_atom_positions_x__bohr>[-+0-9.]+)\s+(?P<x_qbox_geometry_atom_positions_y__bohr>[-+0-9.]+)\s+(?P<x_qbox_geometry_atom_positions_z__bohr>[-+0-9.]+)\s*</position>", repeats = True),
            ], repeats = True)
        ])


    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [

        SM (name = 'Root2',
            startReStr = "",
            sections = ['section_single_configuration_calculation'],
            forwardMatch = True,
            weak = True,
            subMatchers = [

            geometrySubMatcher


    ]), # CLOSING section_single_configuration_calculation
])


def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_run': CachingLvl,
                               'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_qbox_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the qbox *.xml file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections  run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    QboxXMLFileSimpleMatcher = build_QboxXMLFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/qbox.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'qbox-xml-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = QboxXMLFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = QboxXMLParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)





# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from structlog import BoundLogger
from flask_restplus import fields
from datetime import datetime
import pytz
from contextlib import contextmanager

from nomad import config


logger: BoundLogger = None
''' A logger pre configured with information about the current request. '''

base_path = config.services.api_base_path
''' Provides the root path of the nomad APIs. '''


class RFC3339DateTime(fields.DateTime):

    def format(self, value):
        if isinstance(value, datetime):
            return super().format(value.replace(tzinfo=pytz.utc))
        else:
            return str(value)


rfc3339DateTime = RFC3339DateTime()


class DotKeyFieldMixin:
    ''' Allows use of flask_restplus fields with '.' in key names. By default, '.'
    is used as a separator for accessing nested properties. Mixin prevents this,
    allowing fields to use '.' in the key names.

    Example of issue:
    >>> data = {"my.dot.field": 1234}
    >>> model = {"my.dot.field: fields.String}
    >>> marshal(data, model)
    {"my.dot.field:": None}

    flask_restplus tries to fetch values for data['my']['dot']['field'] instead
    of data['my.dot.field'] which is the desired behaviour in this case.
    '''

    def output(self, key, obj, **kwargs):
        transformed_obj = {k.replace(".", "___"): v for k, v in obj.items()}
        transformed_key = key.replace(".", "___")
        # if self.attribute is set and contains '.' super().output() will
        # use '.' as a separator for nested access.
        # -> temporarily set to None to overcome this
        with self.toggle_attribute():
            data = super().output(transformed_key, transformed_obj)
        return data

    @contextmanager
    def toggle_attribute(self):
        ''' Context manager to temporarily set self.attribute to None

        Yields self.attribute before setting to None
        '''
        attribute = self.attribute
        self.attribute = None
        yield attribute
        self.attribute = attribute


class DotKeyNested(DotKeyFieldMixin, fields.Nested):
    pass



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' All generic entry metadata and related classes. '''

from typing import Any
from cachetools import cached, TTLCache
from elasticsearch_dsl import Keyword, Text, analyzer, tokenizer
import ase.data

from nomad import metainfo, config
from nomad.metainfo.search_extension import Search
from nomad.metainfo.elastic_extension import ElasticDocument
from nomad.metainfo.mongoengine_extension import Mongo, MongoDocument
from nomad.datamodel.metainfo.public import fast_access

from .dft import DFTMetadata
from .ems import EMSMetadata
from .qcms import QCMSMetadata

# This is usually defined automatically when the first metainfo definition is evaluated, but
# due to the next imports requireing the m_package already, this would be too late.
m_package = metainfo.Package()

from .encyclopedia import EncyclopediaMetadata  # noqa
from .metainfo.public import section_run, Workflow  # noqa
from .metainfo.common_experimental import Experiment  # noqa
from .metainfo.general_qcms import QuantumCMS  # noqa


def _only_atoms(atoms):
    numbers = [ase.data.atomic_numbers[atom] for atom in atoms]
    only_atoms = [ase.data.chemical_symbols[number] for number in sorted(numbers)]
    return ''.join(only_atoms)


path_analyzer = analyzer(
    'path_analyzer',
    tokenizer=tokenizer('path_tokenizer', 'pattern', pattern='/'))


class Author(metainfo.MSection):
    ''' A person that is author of data in NOMAD or references by NOMAD. '''
    name = metainfo.Quantity(
        type=str,
        derived=lambda user: ('%s %s' % (user.first_name, user.last_name)).strip(),
        a_search=Search(mapping=Text(fields={'keyword': Keyword()})))

    first_name = metainfo.Quantity(type=metainfo.Capitalized)
    last_name = metainfo.Quantity(type=metainfo.Capitalized)
    email = metainfo.Quantity(
        type=str,
        a_elastic=dict(mapping=Keyword),  # TODO remove?
        a_search=Search())

    affiliation = metainfo.Quantity(type=str)
    affiliation_address = metainfo.Quantity(type=str)


class User(Author):
    ''' A NOMAD user.

    Typically a NOMAD user has a NOMAD account. The user related data is managed by
    NOMAD keycloak user-management system. Users are used to denote uploaders, authors,
    people to shared data with embargo with, and owners of datasets.

    Args:
        user_id: The unique, persistent keycloak UUID
        username: The unique, persistent, user chosen username
        first_name: The users first name (including all other given names)
        last_name: The users last name
        affiliation: The name of the company and institutes the user identifies with
        affiliation_address: The address of the given affiliation
        create: The time the account was created
        repo_user_id: The id that was used to identify this user in the NOMAD CoE Repository
        is_admin: Bool that indicated, iff the user the use admin user
    '''

    user_id = metainfo.Quantity(
        type=str,
        a_search=Search())

    username = metainfo.Quantity(type=str)

    created = metainfo.Quantity(type=metainfo.Datetime)

    repo_user_id = metainfo.Quantity(
        type=str,
        description='Optional, legacy user id from the old NOMAD CoE repository.')

    is_admin = metainfo.Quantity(
        type=bool, derived=lambda user: user.user_id == config.services.admin_user_id)

    @staticmethod
    @cached(cache=TTLCache(maxsize=2048, ttl=24 * 3600))
    def get(*args, **kwargs) -> 'User':
        from nomad import infrastructure
        return infrastructure.keycloak.get_user(*args, **kwargs)  # type: ignore


class UserReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use user_ids as values. It automatically
    resolves user_ids to User objects. This is done lazily on getting the value.
    '''

    def __init__(self):
        super().__init__(User.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        return User.get(user_id=proxy.m_proxy_value)

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        return value.user_id


user_reference = UserReference()


class AuthorReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use either user_ids or direct author
    information as values. It automatically resolves user_ids to User objects and author
    data into Author objects.
    '''

    def __init__(self):
        super().__init__(Author.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        proxy_value = proxy.m_proxy_value
        if isinstance(proxy_value, str):
            return User.get(user_id=proxy.m_proxy_value)
        elif isinstance(proxy_value, dict):
            return Author.m_from_dict(proxy_value)
        else:
            raise metainfo.MetainfoReferenceError()

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        if isinstance(value, User):
            return value.user_id
        elif isinstance(value, Author):
            return value.m_to_dict()
        else:
            raise metainfo.MetainfoReferenceError()


author_reference = AuthorReference()


class Dataset(metainfo.MSection):
    ''' A Dataset is attached to one or many entries to form a set of data.

    Args:
        dataset_id: The unique identifier for this dataset as a string. It should be
            a randomly generated UUID, similar to other nomad ids.
        name: The human readable name of the dataset as string. The dataset name must be
            unique for the user.
        user_id: The unique user_id of the owner and creator of this dataset. The owner
            must not change after creation.
        doi: The optional Document Object Identifier (DOI) associated with this dataset.
            Nomad can register DOIs that link back to the respective representation of
            the dataset in the nomad UI. This quantity holds the string representation of
            this DOI. There is only one per dataset. The DOI is just the DOI name, not its
            full URL, e.g. "10.17172/nomad/2019.10.29-1".
        pid: The original NOMAD CoE Repository dataset PID. Old DOIs still reference
            datasets based on this id. Is not used for new datasets.
        created: The date when the dataset was first created.
    '''
    m_def = metainfo.Section(a_mongo=MongoDocument())

    dataset_id = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(primary_key=True),
        a_search=Search())
    name = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True),
        a_search=Search())
    user_id = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True))
    doi = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True),
        a_search=Search())
    pid = metainfo.Quantity(
        type=str,
        a_mongo=Mongo(index=True))
    created = metainfo.Quantity(
        type=metainfo.Datetime,
        a_mongo=Mongo(index=True),
        a_search=Search())


class DatasetReference(metainfo.Reference):
    '''
    Special metainfo reference type that allows to use dataset_ids as values. It automatically
    resolves dataset_ids to Dataset objects. This is done lazily on getting the value.
    '''

    def __init__(self):
        super().__init__(Dataset.m_def)

    def resolve(self, proxy: metainfo.MProxy) -> metainfo.MSection:
        return Dataset.m_def.a_mongo.get(dataset_id=proxy.m_proxy_value)

    def serialize(self, section: metainfo.MSection, quantity_def: metainfo.Quantity, value: Any) -> Any:
        if isinstance(value, metainfo.MProxy):
            return value.m_proxy_value
        else:
            return value.dataset_id


dataset_reference = DatasetReference()


class UserProvidableMetadata(metainfo.MCategory):
    ''' NOMAD entry metadata quantities that can be determined by the user, e.g. via nomad.yaml. '''


class EditableUserMetadata(metainfo.MCategory):
    ''' NOMAD entry metadata quantities that can be edited by the user after publish. '''
    m_def = metainfo.Category(categories=[UserProvidableMetadata])


class MongoMetadata(metainfo.MCategory):
    ''' NOMAD entry quantities that are stored in mongodb and not necessarely in the archive. '''
    pass


class DomainMetadata(metainfo.MCategory):
    ''' NOMAD entry quantities that are determined by the uploaded data. '''
    pass


def derive_origin(entry):
    if entry.external_db is not None:
        return str(entry.external_db)

    if entry.uploader:
        return entry.uploader.name

    return None


def derive_authors(entry):
    uploaders = []
    if entry.uploader is not None and entry.external_db is None:
        uploaders = [entry.uploader]
    return uploaders + entry.coauthors


class EntryMetadata(metainfo.MSection):
    '''
    Attributes:
        upload_id: The ``upload_id`` of the calculations upload (random UUID).
        calc_id: The unique mainfile based calculation id.
        calc_hash: The raw file content based checksum/hash of this calculation.
        pid: The unique persistent id of this calculation.
        mainfile: The upload relative mainfile path.
        domain: Must be the key for a registered domain. This determines which actual
            subclass is instantiated.

        files: A list of all files, relative to upload.
        upload_time: The time when the calc was uploaded.
        uploader: An object describing the uploading user, has at least ``user_id``
        processed: Boolean indicating if this calc was successfully processed and archive
            data and calc metadata is available.
        last_processing: A datatime with the time of the last successful processing.
        nomad_version: A string that describes the version of the nomad software that was
            used to do the last successful processing.

        comment: An arbitrary string with user provided information about the entry.
        references: A list of URLs for resources that are related to the entry.
        uploader: Id of the uploader of this entry.
        coauthors: Ids of all co-authors (excl. the uploader) of this entry. Co-authors are
            shown as authors of this entry alongside its uploader.
        shared_with: Ids of all users that this entry is shared with. These users can find,
            see, and download all data for this entry, even if it is in staging or
            has an embargo.
        with_embargo: Entries with embargo are only visible to the uploader, the admin
            user, and users the entry is shared with (see shared_with).
        upload_time: The time that this entry was uploaded
        datasets: Ids of all datasets that this entry appears in
    '''
    m_def = metainfo.Section(
        a_elastic=ElasticDocument(index_name=config.elastic.index_name, id=lambda x: x.calc_id))

    upload_id = metainfo.Quantity(
        type=str,
        description='The persistent and globally unique identifier for the upload of the entry',
        a_search=Search(
            many_or='append', group='uploads_grouped', metric_name='uploads', metric='cardinality'))

    calc_id = metainfo.Quantity(
        type=str,
        description='A persistent and globally unique identifier for the entry',
        a_search=Search(many_or='append'))

    calc_hash = metainfo.Quantity(
        type=str,
        description='A raw file content based checksum/hash',
        categories=[MongoMetadata],
        a_search=Search(
            many_or='append', metric_name='unique_entries', metric='cardinality'))

    mainfile = metainfo.Quantity(
        type=str,
        description='The path to the mainfile from the root directory of the uploaded files',
        a_search=[
            Search(
                description='Search within the mainfile path.',
                mapping=Text(multi=True, analyzer=path_analyzer, fields={'keyword': Keyword()}),
                many_or='append', search_field='mainfile.keyword'),
            Search(
                description='Search for the exact mainfile.',
                many_and='append', name='mainfile_path', search_field='mainfile.keyword')])

    files = metainfo.Quantity(
        type=str, shape=['0..*'],
        description='''
        The paths to the files within the upload that belong to this entry.
        All files within the same directory as the entry's mainfile are considered the
        auxiliary files that belong to the entry.
        ''',
        a_search=[
            Search(
                description='Search within the paths.', name='path',
                mapping=Text(
                    multi=True, analyzer=path_analyzer, fields={'keyword': Keyword()})
            ),
            Search(
                description='Search for exact paths.',
                many_or='append', name='files', search_field='files.keyword')])

    pid = metainfo.Quantity(
        type=str,
        description='''
        The unique, sequentially enumerated, integer PID that was used in the legacy
        NOMAD CoE. It allows to resolve URLs of the old NOMAD CoE Repository.''',
        categories=[MongoMetadata],
        a_search=Search(many_or='append'))

    raw_id = metainfo.Quantity(
        type=str,
        description='''
        The code specific identifier extracted from the entrie's raw files if such an
        identifier is supported by the underlying code
        ''',
        categories=[MongoMetadata, UserProvidableMetadata],
        a_search=Search(many_or='append'))

    domain = metainfo.Quantity(
        type=metainfo.MEnum('dft', 'ems', 'qcms'),
        description='The material science domain',
        categories=[MongoMetadata, UserProvidableMetadata],
        a_search=Search())

    published = metainfo.Quantity(
        type=bool, default=False,
        description='Indicates if the entry is published',
        categories=[MongoMetadata],
        a_search=Search())

    processed = metainfo.Quantity(
        type=bool, default=False,
        description='Indicates that the entry is successfully processed.',
        categories=[MongoMetadata],
        a_search=Search())

    last_processing = metainfo.Quantity(
        type=metainfo.Datetime,
        description='The datetime of the last processing',
        categories=[MongoMetadata],
        a_search=Search())

    nomad_version = metainfo.Quantity(
        type=str,
        description='The NOMAD version used for the last processing',
        categories=[MongoMetadata],
        a_search=Search(many_or='append'))
    nomad_commit = metainfo.Quantity(
        type=str,
        description='The NOMAD commit used for the last processing',
        categories=[MongoMetadata],
        a_search=Search(many_or='append'))
    parser_name = metainfo.Quantity(
        type=str,
        description='The NOMAD parser used for the last processing',
        a_search=Search(many_or='append'))

    comment = metainfo.Quantity(
        type=str, categories=[MongoMetadata, EditableUserMetadata],
        description='A user provided comment for this entry',
        a_search=Search(mapping=Text()))

    references = metainfo.Quantity(
        type=str, shape=['0..*'], categories=[MongoMetadata, EditableUserMetadata],
        description='User provided references (URLs) for this entry',
        a_search=Search())

    external_db = metainfo.Quantity(
        type=metainfo.MEnum('EELSDB'), categories=[MongoMetadata, UserProvidableMetadata],
        description='The repository or external database where the original data resides',
        a_search=Search())

    uploader = metainfo.Quantity(
        type=user_reference, categories=[MongoMetadata],
        description='The uploader of the entry',
        a_flask=dict(admin_only=True, verify=User),
        a_search=[
            Search(
                description='The full name of the authors for exact searches',
                metric_name='uploaders', metric='cardinality',
                many_or='append', search_field='uploader.name.keyword',
                statistic_size=10,
                statistic_order='_count'),
            Search(
                name='uploader_id', search_field='uploader.user_id',
                description='The full name of the authors',)
        ])

    origin = metainfo.Quantity(
        type=str,
        description='''
            A short human readable description of the entries origin. Usually it is the
            handle of an external database/repository or the name of the uploader.
        ''',
        derived=derive_origin,
        a_search=Search(statistic_size=10, statistic_order='_count'))

    coauthors = metainfo.Quantity(
        type=author_reference, shape=['0..*'], default=[], categories=[MongoMetadata, EditableUserMetadata],
        description='A user provided list of co-authors',
        a_flask=dict(verify=User))

    authors = metainfo.Quantity(
        type=author_reference, shape=['0..*'],
        description='All authors (uploader and co-authors)',
        derived=derive_authors,
        a_search=Search(
            description='The full name of the authors for exact searches',
            metric='cardinality',
            many_or='append', search_field='authors.name.keyword'))

    shared_with = metainfo.Quantity(
        type=user_reference, shape=['0..*'], default=[], categories=[MongoMetadata, EditableUserMetadata],
        description='A user provided list of userts to share the entry with',
        a_flask=dict(verify=User))

    owners = metainfo.Quantity(
        type=user_reference, shape=['0..*'],
        description='All owner (uploader and shared with users)',
        derived=lambda entry: ([entry.uploader] if entry.uploader is not None else []) + entry.shared_with,
        a_search=Search(
            description='The full name of the owners for exact searches',
            many_or='append', search_field='owners.name.keyword'))

    with_embargo = metainfo.Quantity(
        type=bool, default=False, categories=[MongoMetadata, EditableUserMetadata],
        description='Indicated if this entry is under an embargo',
        a_search=Search())

    upload_time = metainfo.Quantity(
        type=metainfo.Datetime, categories=[MongoMetadata],
        description='The date and time this entry was uploaded to nomad',
        a_flask=dict(admin_only=True),
        a_search=Search(order_default=True))

    upload_name = metainfo.Quantity(
        type=str, categories=[MongoMetadata],
        description='The user provided upload name',
        a_search=Search(many_or='append'))

    datasets = metainfo.Quantity(
        type=dataset_reference, shape=['0..*'], default=[],
        categories=[MongoMetadata, EditableUserMetadata],
        description='A list of user curated datasets this entry belongs to.',
        a_flask=dict(verify=Dataset),
        a_search=[
            Search(
                search_field='datasets.name', many_or='append',
                description='A list of user curated datasets this entry belongs to for exact name search'),
            Search(
                name='dataset_id', search_field='datasets.dataset_id', many_or='append',
                group='datasets_grouped',
                metric='cardinality', metric_name='datasets',
                description='A list of user curated datasets this entry belongs to for exact name search')])

    external_id = metainfo.Quantity(
        type=str, categories=[MongoMetadata, UserProvidableMetadata],
        description='''
        A user provided external id. Usually the id for an entry in an external database
        where the data was imported from.''',
        a_search=Search(many_or='split'))

    last_edit = metainfo.Quantity(
        type=metainfo.Datetime, categories=[MongoMetadata],
        description='The date and time the user metadata was edited last',
        a_search=Search())

    formula = metainfo.Quantity(
        type=str, categories=[DomainMetadata],
        description='A (reduced) chemical formula',
        a_search=Search())

    atoms = metainfo.Quantity(
        type=str, shape=['n_atoms'], default=[], categories=[DomainMetadata],
        description='The atom labels of all atoms of the entry\'s material',
        a_search=Search(
            many_and='append', statistic_size=len(ase.data.chemical_symbols)))

    only_atoms = metainfo.Quantity(
        type=str, categories=[DomainMetadata],
        description='The atom labels concatenated in order-number order',
        derived=lambda entry: _only_atoms(entry.atoms),
        a_search=Search(many_and='append', derived=_only_atoms))

    n_atoms = metainfo.Quantity(
        type=int, categories=[DomainMetadata], default=0,
        description='The number of atoms in the entry\'s material',
        a_search=Search())

    ems = metainfo.SubSection(sub_section=EMSMetadata, a_search='ems')
    dft = metainfo.SubSection(sub_section=DFTMetadata, a_search='dft', categories=[fast_access])
    qcms = metainfo.SubSection(sub_section=QCMSMetadata, a_search='qcms')
    encyclopedia = metainfo.SubSection(sub_section=EncyclopediaMetadata, a_search='encyclopedia')

    def apply_user_metadata(self, metadata: dict):
        ''' Applies a user provided metadata dict to this calc. '''
        self.m_update(**metadata)

    def apply_domain_metadata(self, archive):
        ''' Used to apply metadata that is related to the domain. '''
        assert self.domain is not None, 'all entries must have a domain'
        domain_sub_section_def = self.m_def.all_sub_sections.get(self.domain)
        domain_section_def = domain_sub_section_def.sub_section
        assert domain_section_def is not None, 'unknown domain %s' % self.domain

        # add domain section if not already there
        domain_section = self.m_get_sub_section(domain_sub_section_def, -1)
        if domain_section is None:
            domain_section = self.m_create(domain_section_def.section_cls)

        domain_section.apply_domain_metadata(archive)


class EntryArchive(metainfo.MSection):

    section_run = metainfo.SubSection(sub_section=section_run, repeats=True)
    section_experiment = metainfo.SubSection(sub_section=Experiment)
    section_quantum_cms = metainfo.SubSection(sub_section=QuantumCMS)
    section_workflow = metainfo.SubSection(sub_section=Workflow, categories=[fast_access])
    section_metadata = metainfo.SubSection(sub_section=EntryMetadata, categories=[fast_access])

    processing_logs = metainfo.Quantity(
        type=Any, shape=['0..*'],
        description='The processing logs for this entry as a list of structlog entries.')
